#!/itl/links/generic/bin/tclsh8.0

# General Configuration
set debug 0
set workDir [pwd]
set cgiPath "/sasp/cgi.tcl-0.7"
set wsRoot "/sasp/WebSubmit"
set publicDir "$wsRoot/public"
set distDir "$wsRoot/webSubmit2.5"
set DEV(public) $publicDir
set DEV(public,user) "nobody"
set DEV(public,group) websbmt
set DEV(dist) $distDir
set DEV(dist,user) "nobody"
set DEV(dist,group) websbmt
set sccsDir "$wsRoot/develop/src"
set distFile "$sccsDir/websubmit.dist"
set serverUser "nobody"
set serverGroup "websbmt"
set serverDirList [list "user" "lib/LL"]

# Developer directory information
set devUsers [list koontz ryanm judy]
foreach usr $devUsers {
    set DEV($usr) "$wsRoot/develop/dev$usr"
    set DEV($usr,user) $usr
    set DEV($usr,group) websbmt
}

# Change umask to avoid permissions issues
if [catch {exec /sbin/umask 000} err] {
    puts "WARNING: Could not change umask ($err)"
}

# Variables that need updating in source code
set fixVarList [list wsRootDir cgiPath serverBinAlias nServerBinAlias wsLibDir nServerDocAlias]

# Variable values in developer trees
foreach usr $devUsers {
    set FIXVAR(wsRootDir,$usr) $DEV($usr)
    set FIXVAR(wsLibDir,$usr) [file join $DEV($usr) lib]
    set FIXVAR(cgiPath,$usr) $cgiPath
    set FIXVAR(serverBinAlias,$usr) dev$usr
    set FIXVAR(nServerBinAlias,$usr) dev$usr
    set FIXVAR(nServerDocAlias,$usr) doc$usr
}

# Variable values in public and distribution directories
foreach dir [list public dist] {
    set FIXVAR(wsRootDir,$dir) $DEV($dir)
    set FIXVAR(wsLibDir,$dir) [file join $DEV($dir) lib]
    set FIXVAR(cgiPath,$dir) $cgiPath
    set FIXVAR(serverBinAlias,$dir) websubmit
    set FIXVAR(nServerBinAlias,$dir) websubmit
    set FIXVAR(nServerDocAlias,$dir) doc
}



#########################################################
#
#                    PROCEDURES
#
#########################################################

################################################
# CODEINFO BEGIN
################################################
# info proc setVarValues> Procedure to set up variable
#     values based on username.  This works for developers
#     as well as public and dist.
proc setVarValues {user} {
    global fixVarList FIXVAR VARS

    # Update variable values
    foreach var $fixVarList {
	if [info exists FIXVAR($var,$user)] {
	    set VARS($var) $FIXVAR($var,$user)
	} else {
	    return -code error "Could not determine value of $var for user $user"
	}
    }
}



################################################
# CODEINFO BEGIN
################################################
# Procedure to determine currently checked out files
proc getSccsInfo {user} {
    global sccsDir
    set pwd [pwd]
    cd $sccsDir

    set chkdFiles ""
    set chkList ""
    if [catch {glob SCCS/p.*} chkdFiles] {
	# No files checked out
    } else {
	foreach file $chkdFiles {
	    if {[file attributes $file -owner] == $user} {
		lappend chkList [string range $file 7 end]
	    }
	}
    }
    cd $pwd
    return $chkList
}


################################################
# CODEINFO BEGIN
################################################
# Procedure to recursively copy all files from one directory to another
#  A set of 'protected' files will not be copied
proc safeCopy {src target protectList {dirList ""}} {
    set pwd [pwd]
    if [catch {cd $src} err] {
	puts stderr $err
	return
    }

    # Copy files
    foreach file [glob -nocomplain *] {
	if ![file isdirectory $file] {
	    if {[lsearch -exact $protectList $file] == -1} {
		# File not protected
		set srcFile [file join $src $file]
		set targFile [file join $target $file] 
		if [catch {file copy -force $file $targFile} err] {
		    puts "\n ERROR (safeCopy): $err\n"
		} else {
		    puts " $srcFile --> $targFile"
		}
	    } else {
		puts "\nWARNING: $file is sccs protected...Not copied\n"
	    }
	}
    }

    foreach file [glob -nocomplain *] {
	if [file isdirectory $file] {
	    if {($dirList == "") || ([lsearch -exact $dirList $file] != -1)} {
		set newTarget [file join $target $file]
		if ![file exists $newTarget] {
		    file mkdir $newTarget
		    safeCopy [file join $src $file] $newTarget $protectList
		    puts " Directory created: $newTarget"
		} else {
		    if ![file isdirectory $newTarget] {
			puts "\nWARNING! Source tree at $file not copied.\n"
		    } else {
			safeCopy [file join $src $file] $newTarget $protectList
		    }
		}
	    }
	}
    }
    cd $pwd
}

################################################
# CODEINFO BEGIN
################################################
# info> Procedure to make a relative pathname absolute
# arg path> Path to fix
# arg root> Root to prepend to path if it is relative
# return> Absolute path corresponding to input
# loc-sclr path> Path to fix
proc mkPathAbs {path root} {
    # Strip trailing slash from root path
    set root [string trimright $root /]

    # Verify that root is in fact an absolute head
    if ![regexp {^/.*} $root match] {
        return $path
    }

    # Check path
    if {[file pathtype $path] == "relative"} {
        set path [join [list $root $path] /]
    }
    return $path
}
  

################################################
# CODEINFO BEGIN
################################################ 
# info > Procedure to obtain the list of files in the WebSubmit distribution
# arg distFile> Name of distribution file (absolute pathname)
# return> List of distribution files or ERROR
# loc-sclr distFile> Name of distribution file (absolute pathname)
# loc-sclr fileId> Filehandle for distribution file
proc getDist {distFile} {
    global DIST VARS distList user

    # Open file containing distribution
    if [catch {open $distFile r} fileId] {
	# Error opening distribution file
	return -code error "$fileId"
    }

    # Read input file
    set distList ""
    while {[gets $fileId line] >= 0} {
	if {[regexp {^#.*} $line match] || ($line == "")} {
	    # Comment or blank line - ignore
	} else {
	    if [regexp {([^ ]+) +([^ ]+)} $line match mode file] {
		# Resolve directory and name components of file
		set workDir [file dirname $file]
		set fileName [file tail $file]
		set DIST($fileName,mode) $mode
		set DIST($fileName,dir) $workDir
		lappend distList $fileName
	    }
	}
    }
    close $fileId
    return 0
}

################################################
# CODEINFO BEGIN
################################################ 
# info > Procedure to delete concrete equivalents A of A.example files
#     from the list of distribution files.  Used with outside distributions.
#     Some *.example files have no concrete equivalent - we never want to
#     retrieve A.
proc filterDist {} {
    global distList

    puts "Removing Concrete Instances of *example files."
    puts "----------------------------------------------"

    foreach d $distList {
	if [regexp "^(.+)\.example$" $d matchAll matchHead] {
	    set matchLoc [lsearch -exact $distList $matchHead]
	    if {$matchLoc != -1} {
		set len1 [llength $distList]
		set distList [lreplace $distList $matchLoc $matchLoc]
		set len2 [llength $distList]
		puts "Removed $matchHead"
	    }		
	}
    }
    return 0
}

################################################
# CODEINFO BEGIN
################################################
# info> Procedure to perform an SCCS get on a group of files
# arg targetDir> Target directory
# arg fileList> List of files in distribution (relative pathnames)
# return>
# loc-sclr targetDir> Target directory
# loc-list fileList> List of files in distribution (relative pathnames)
proc sccsGetDist {targetDir fileList} {
    global DIST sccsDir user DEV distDir
    set srcDir $sccsDir

    puts "\nCopying distribution to $targetDir"
    puts "======================================"
    if {[pwd] != $srcDir} {
	cd $srcDir
    }

    # Verification test for bin directory
    set dirUser $DEV($user,user)
    set dirGroup $DEV($user,group)
    if {![file exists $targetDir]} {
	# Create target directory if possible
	if [catch {file mkdir $targetDir} result] {
	    # Could not create directory
	    error "sccsGetDist: $result"
	} else {
	    if [catch {exec /usr/bin/chmod -R 775 $targetDir} err] {
		puts "WARNING(sccsGetDist): $err"
	    } else {
		if [catch {exec /usr/bin/chown ${dirUser}.${dirGroup} $targetDir} err] {
		    error "sccsGetDist: $err"
		}
	    }
	}
    } else {
	if {![file isdirectory $targetDir] || ![file writable $targetDir]} {
	    # User does not have write permission for targetDir
	    error "sccsGetDist: Specified bin directory is either not a directory or\
		    is not writable"
	}
    }

    foreach file $fileList {
	set targetFile [file join $targetDir $DIST($file,dir) $file]
	set targetFileParent [file dirname $targetFile]
	set srcFile [file tail $file]

	# Create parent directory for target file if it does not exist
	if ![file exists $targetFileParent] {
	    if [catch {file mkdir $targetFileParent} result] {
		error "sccsGetDist: Could not create directory $targetFileParent ($result)"
	    } else {
		if [catch {exec /usr/bin/chmod -R 775 $targetFileParent} err] {
		    puts "WARNING(sccsGetDist): $err"
		} else {
		    if [catch {exec /usr/bin/chown ${dirUser}.${dirGroup} $targetFileParent} err] {
			error "sccsGetDist: $err"
		    }
		}
	    }
	}
	
	switch -regexp [file extension $srcFile] {
	    "^\.(gif|jpg)$" {
	    # Binary files not maintained in SCCS format
		if ![file exists ../doc/$srcFile] {
		    error "sccsGetDist: Could not find image file \"$srcFile\ in \"[pwd]\"\n)"
		}

		# Copy non-sccs file to public directory
		file copy -force ../doc/$srcFile $targetFile
		puts "  $srcFile --> $targetFile"
	    }

	    default {
            # Text files maintained in SCCS format
		if {[file extension $srcFile] == ".example"
		    && "$targetDir" != "$distDir"} {
		    puts "  *** Skip copying example file $targetFile to $targetDir."
		} else {

		    # Get sccs file out of SCCS
		    if [catch {exec sccs get $srcFile} err] {
			error "sccsGetDist: Could not get SCCS file \"$srcFile\"\n ($err)"
		    }
		
		    # Copy sccs file to public directory
		    file copy -force $srcFile $targetFile
		    puts "  $srcFile --> $targetFile"
		    file delete $srcFile
		    
		    # If file is a java file compile it
		    if {[file extension $srcFile] == ".java"} {
			puts "  *** $targetFile is a java file and needs compiling."
			set javaWorkingDir [file join $targetDir $DIST($file,dir)]
			set classFileList [glob -nocomplain $javaWorkingDir/*.class]
			catch {eval exec /sbin/rm  $classFileList}
			if [catch {exec /usr/java/bin/javac -classpath /itl/apps/netscape-4.61/sparc.sun.solaris2.5.1/java/classes/java40.jar $targetFile} err] {
			    error "sccsGetDist: Could not compile Java file \"$srcFile\"\n ($err)"
			}
			set classFileList [glob -nocomplain $javaWorkingDir/*.class]
			if [catch {eval exec /usr/bin/chmod -R 664 $classFileList} err] {
			    puts "WARNING(sccsGetDist): $err"
			} else {
			    if [catch {eval exec /usr/bin/chown ${dirUser}.${dirGroup} $classFileList} err] {
				error "sccsGetDist: $err"
			    }
			}		
		    }
		}		    
	    }
	}
    }
    return 0
}


############################################
# CODEINFO BEGIN
############################################
# info proc fixSrc> Procedure to fix source
#   commands in a group of files
proc fixSrc {wsRootDir} {
    global VARS

    # Fix source commands in bin directories
    puts -nonewline "\nUpdating source commands..."
#   set wsRootDir "/sasp/WebSubmit/webSubmit2.5"
    foreach file [glob $wsRootDir/bin/*.cgi] {
	updateTclSource $file websubmit.conf [file join $wsRootDir bin]
    }
    puts "done"
}


############################################
# CODEINFO BEGIN
############################################
# info proc fileUpdate> Update variables in a group of files
proc fileUpdate {fileList varList} {
    global UPDATEVARS DIST
    catch {unset UPDATEVARS}

    puts "\nUpdating variables..."
    # Get full path names for files
    foreach file $fileList {
	lappend newFileList $DIST($file,fullPath)
    }

    # Determine which files in list contain variables that need changing
    foreach var $varList {
	if ![catch {eval exec /sbin/grep -l $var $newFileList} foundList] {
	    foreach file $foundList {
		lappend UPDATEVARS($file) $var
	    }
	}
    }

    # Update necessary files
    foreach file [array names UPDATEVARS] {
	# Create a backup copy of file
	file copy -force $file $file.bak

	# Update variables
	if [catch {updateVars $file} error] {
	    file copy -force $file.bak $file
	    puts "Problem updating $file - changes discarded\nError Message: $error"
	    return -code return
	} else {
	    # Remove backup file
	    file delete $file.bak
	}
    }
}


################################################
# CODEINFO BEGIN
################################################
# info> Procedure to update Tcl source commands so that the sourced files
#     contain the proper full pathnames
# arg srcFile> File to scan for 'source' commands
# arg fileName> Filename used in source command
# arg dirName> Path to filename
# return> None

proc updateTclSource {srcFile fileName dirName} {
    global debug

    # Certain files are marked for no update
    set fileDir [file dirname $srcFile]
    if [file exists [file join $fileDir noupdate]] {
	# Don't update files in this directory!!!
	puts "Files not updated in $fileDir"
	return
    }

    # Open file, scan for entry, and replace as necessary
    if [catch {open $srcFile r+} fileId] {
	return -code error "updateTclSource: $fileId"
    }

    # Extract lines and store
    set lines [split [read $fileId] \n]
    close $fileId

    # Reopen file and truncate
    if [catch {open $srcFile w} fileId] {
	return -code error $fileId
    }    
    
    # Find line that matches source command
    set srcId [lsearch -regexp $lines " *source .*$fileName"]
    set fullName [file join $dirName $fileName]
    set white ""
    if {$srcId != -1} {
	regexp "(\[\#\; \t\]*)source .*$fileName\$" $srcId match white 
	set lines [lreplace $lines $srcId $srcId "${white}source $fullName"]
	if {$debug} {
	    puts " $srcFile --> ${white}source $fullName"
	}
    }

    # Write data
    foreach line $lines {
	puts $fileId $line
    }
    close $fileId
    return 0
}


################################################
# CODEINFO BEGIN
################################################
# info> Procedure to scan a file and replace specified variable
#  definitions with new values.  Not done in the most efficient
#  manner, but it works...

# arg file> Name of file to update
# return> Error status
# loc-sclr file> Name of file to update
# loc-sclr fileId> Filehandle for File   
proc updateVars {file} {
    global UPDATEVARS VARS debug

    if {$debug} {
	puts "Updating variables in $file"
    }

    # Certain files are marked for no update
    set fileDir [file dirname $file]
    if [file exists [file join $fileDir noupdate]] {
	# Don't update files in this directory!!!
	return
    }

    # Open file, scan for entry, and replace as necessary
    set fileName [file tail $file]
    if [catch {open $file r+} fileId] {
	return -code error "Could not read $file for updating:\n$fileId"
    }

    # Extract lines and store
    set lines [split [read $fileId] \n]
    close $fileId

    # Reopen file and truncate
    if [catch {open $file w} fileId] {
	# Restore backup copy
	file copy -force $file.bak $file
	return -code error "Could not update $file:\n$fileId"
    }

    # Variable definitions
    foreach var $UPDATEVARS($file) {
	if ![regexp {([^\(]+)\((.*)\)} $var match arrName key] {
	    # Variable is string or list
	    global $var
	    set TYPE($var) "scalar"
	} else {
	    # Variable is array
	    global $arrName
	    set TYPE($var) "array"
	    set TYPE($var,name) $arrName
	    set TYPE($var,key) $key
	}
    }

    foreach line $lines {
	set outputLine $line
	foreach var $UPDATEVARS($file) {
	    if {$TYPE($var) == "scalar"} {
		set value $VARS($var)
		if [regexp "(\[\#\; \t\]*)(set|variable|readOnly) +$var +(\[^ \]+)" $line match white varName tail] {
		    # Found set expression for variable
		    puts "    $fileName: $var ---> $value"
		    set outputLine "$white$varName $var \"$value\""
		    break
		}
	    } else {
		# Variable is array
		set value [set $var]
		set arrName $TYPE($var,name)
		set key $TYPE($var,key)
		if [regexp "(\[\#\; \t\]*)readOnlyArray +$arrName +$key +\[^ \]+ (.*)" $line match white tail] {
		    # Found readOnlyArray expression for variable
		    puts "    $fileName: $var ---> $value"
		    set outputLine "${white}readOnlyArray $arrName $key \"$value\" $tail"
		    break
		} elseif [regexp "(\[\#\; \t\]*)set +$var +(\[^ \]+)" $line match white] {
		    # Found readOnlyArray expression for variable
		    puts "    $fileName: $var ---> $value"
		    set outputLine "${white}set $var \"$value\""
		    break
		}
	    }
	}
	puts $fileId $outputLine
    }
    close $fileId
    return 0
}

################################################
# CODEINFO BEGIN
################################################
# Procedure to fix permissions on a set of files
proc fixPerms {srcDir fileList} {
    global DIST workDir
    set PERM(.cgi) 0775
    set PERM(.tcl) 0665
    set PERM(.db) 0665
    puts -nonewline "\nChanging file permissions..."
    foreach file $fileList {
	if {[info exists DIST($file,mode)]} {
	    set fileName [file join $srcDir $DIST($file,dir) $file]
	    if [file exists "$fileName"] {
		if [catch {exec /usr/bin/chmod $DIST($file,mode) "$fileName"} result] {
		    puts "WARNING (chmod): $result"
		}
	    }
	} else {
	    # Set up default permissions based on extension and assume that file is
	    #  in working directory
	    #    *.cgi: 0775
	    #    *.tcl: 0665
	    #    *.???: 0664
	    set fileName [file join $workDir $file]
	    if [file exists "$fileName"] {
		set ext [file extension $file]
		if [info exists PERM($ext)] {
		    set perm $PERM($ext) 
		} else {
		    set perm 0664
		}
		if [catch {exec /usr/bin/chmod $perm "$fileName"} result] {
		    puts "WARNING (fixperms): $result"
		}
	    } else {
		puts "\nERROR (fixPerms): $file does not exist"
	    }
	}
    }
    puts "done"
}

################################################
# CODEINFO BEGIN
################################################
# Procedure to fix ownership on a set of files
proc fixOwn {srcDir fileList user group} {
    global DIST workDir

    puts "\nChanging file ownership..."
    foreach file $fileList {
	set fileName [file join $srcDir $DIST($file,dir) $file]
	if [file exists "$fileName"] {
	    # Change user and group ownership
	    if [catch {exec /sbin/chown ${user}.$group "$fileName"} result] {
		puts "WARNING (chown): $result"
	    }
	}
    }
}

################################################
# CODEINFO BEGIN
################################################
# info> Procedure to determine full path names of files in distribution
proc setDistFileNames {user} {
    global DIST FIXVAR workDir distList

    # Determine if file is in distribution and create appropriate
    #  pathname if it is
    foreach file $distList {
	if [info exists DIST($file,dir)] {
	    set DIST($file,fullPath) [file join $FIXVAR(wsRootDir,$user) $DIST($file,dir) $file]
	} else {
	    set DIST($file,fullPath) [file join $workDir $file]
	}
    }
}

################################################
# CODEINFO BEGIN
################################################
# info proc checkDist> Procedure to check the permissions and
#     ownerships of files within a specified directory tree
proc checkDist {user {fix ""}} {
    global DIST FIXVAR distList DEV dirList serverDirList serverUser serverGroup
    set wsRootDir "/sasp/WebSubmit/webSubmit2.5"
    set userName $DEV($user,user)
    set group $DEV($user,group)
    set FERR() {}
    set DERR() {}
    set MISSING() {}

    # Check for all distribution files
    foreach file $distList {
	set fName $DIST($file,fullPath)
	if ![file exists $fName] {
	    set MISSING($fName) 1
	}
    }

    # Perform tests for all directories
    getDirs $wsRootDir
    foreach dir $dirList {
	# Determine if this is a directory owned by the server user
	regexp "$wsRootDir/(.*)" $dir match dirName
	set serverDir 0
	foreach sd $serverDirList {
	    if [string match "$sd*" $dirName] {
		set serverDir 1
	    }
	}
	set DERR($dir) 0
	set dUser [file attributes $dir -owner]
	set dGroup [file attributes $dir -group]
	set dPerms [file attributes $dir -permissions]
	regexp {.*([0-9][0-9][0-9])$} $dPerms match dPerms

	if {$serverDir} {
	    if {$dUser != "$serverUser"} {
		# Error in user ownership
		set DERR($dir) 1
		set DERR($dir,u) "$dUser ($serverUser)*"
	    }
	} else {
	    if {$dUser != "$userName"  && $dUser != $serverUser} {
		# Error in user ownership
		set DERR($dir) 1
		set DERR($dir,u) "$dUser ($userName)*"
	    }
	}

	if {$dGroup != "$group"} {
	    # Error in group ownership
	    set DERR($dir) 1
	    set DERR($dir,g) "$dGroup ($group)"
	}

	if {$dPerms != "775" && $dPerms != "755" } {
	    # Error in permissions
	    set DERR($dir) 1
	    set DERR($dir,p) "$dPerms (775|755)*"
	}
    }

    # Perform tests for all files
    foreach file $distList {
	set FERR($file) 0
	set fName $DIST($file,fullPath)
	if [info exists MISSING($fName)] {
	    continue
	}

	set distPerms $DIST($file,mode)
	set fUser [file attributes $fName -owner]
	set fGroup [file attributes $fName -group]
	set fPerms [file attributes $fName -permissions]
	regexp {.*([0-9][0-9][0-9])$} $fPerms match fPerms

	if {$fUser != "$userName" && $fUser != $serverUser} {
	    # Error in user ownership
	    set FERR($file) 1
	    set FERR($file,u) "$fUser ($userName)*"
	}

	if {$fGroup != "$group"} {
	    # Error in group ownership
	    set FERR($file) 1
	    set FERR($file,g) "$fGroup ($group)"
	}

	if {$fPerms != "$distPerms"} {
	    # Error in file permissions
	    set FERR($file) 1
	    set FERR($file,p) "$fPerms ($distPerms)"
	}
    }

    # Generate report specifying problems

    puts "\n\nDISTRIBUTION STATUS: USER = $user"
    puts "--------------------------------------"
    puts "Results are listed by file; erroneous values are given, along\n\
	    with the correct value in parentheses.  A '-' indicates that\n\
	    the file has the proper values for that field.  Note that not all\n\
	    problems are necessarily destructive.  Fields marked with an asterisk\n\
	    should be changed."

    puts "\n\nFILES MISSING FROM CURRENT TREE:"
    puts "--------------------------------"
    foreach file [array names MISSING] {
	if {$file != ""} {
	    puts "  $file"
	}
    }

    set header [format "%-45s %-16s %-20s %-9s" DIRECTORY USER GROUP PERMS]
    puts "\n\n$header"
    puts "---------------------------------------------------------------------------------------"
    foreach dir $dirList {
	if {$DERR($dir)} {
	    # Errors found
	    foreach err [list u g p] {
		if [info exists DERR($dir,$err)] {
		    set $err $DERR($dir,$err)
		} else {
		    set $err "-"
		}
	    }
	    regexp "$wsRootDir/(.*)" $dir match dirName
	    set string [format "%-45s %-16s %-20s %-9s" $dirName $u $g $p]
	    puts "$string"
	}
    }


    set header [format "%-25s %-20s %-20s %-9s" FILE USER GROUP PERMS]
    puts "\n\n$header"
    puts "---------------------------------------------------------------------------"
    foreach file $distList {
	if {$FERR($file)} {
	    # Errors found
	    foreach err [list u g p] {
		if [info exists FERR($file,$err)] {
		    set $err $FERR($file,$err)
		} else {
		    set $err "-"
		}
	    }
	    set string [format "%-25s %-20s %-20s %-9s" $file $u $g $p]
	    puts "$string"
	}
    }
}

###########################################
# CODEINFO BEGIN
###########################################
# info proc getDirs> Procedure to recursively descend a directory
#     hierarchy and return a list of all the directories
proc getDirs {rootDir} {
    global dirList
    set pwd [pwd]
    if [catch {cd $rootDir} err] {
	return
    }

    foreach match [glob -nocomplain *] {
	if [file isdirectory $match] {
	    set dirName [file join $rootDir $match]
	    lappend dirList $dirName
	    getDirs $dirName
	}
    }
    cd $pwd
}


###############################################################################
# Create developer's Release
###############################################################################

# Determine current user and validate
if [catch {exec whoami} user] {
    error "Error determining script user"
} else {
    # Check against list of valid users
    if {[lsearch -exact $devUsers $user] == -1} {
	error "Current user ($user) is not allowed to access WebSubmit development scripts"
    }
}



# Parse argument list
set makePublic 0
set makeOutside 0
set copy 0
set update 0
set showInfo 0
set fixdev 0
set fixpub 0
set fixdist 0
set checkPub 0
set checkDev 0
set sccsFileList ""
set tmpFileList ""
set tmpSccsFileList ""
set sccsCmd ""
set sccsOpts ""
set numOpts 0
if {[llength $argv] == 0} {
    puts "\nwebsccs: SCCS control of WebSubmit distribution"
    puts "-----------------------------------------------"
    puts "  USAGE: websccs \[flags\] \[options\] \[filenames\]\n"
    puts "  FLAGS (specify only one):"
    puts "    -fixdev: Update variable names in developer's distribution.  Exit after update.\n"
    puts "    -fixpub: Update variable names in public code.  Exit after update.\n"
    puts "    -fixdist: Update variable names in distribution code.  Exit after update.\n"
    puts "    -public: Create public copy of distribution.  Program exits after\n\
	               \t\tdistribution copied if no filenames given.  If filenames\n\
                       \t\tare given, then only these files will be updated.\n"
    puts "    -outside: Create outside copy of distribution.  Program exits after\n\
	               \t\tdistribution copied if no filenames given.  If filenames\n\
                       \t\tare given, then only these files will be updated.\n"
    puts "    -update: Update developer directory with distribution.  Exit after update.\n"
    puts "    -checkPub: Verify public distribution (existence, ownership, permissions).\n"
    puts "    -checkDev: Verify developer distribution (existence, ownership, permissions).\n"
    puts "    command: A valid SCCS command from the set (info, clean, print, get, edit, delget,\n\
	               \t\tdelta, unedit, enter, create).\n"
    puts "  OPTIONS:"
    puts "   SCCS options for the specified command.  NOTE: If no -y option is passed\n\
	    \tto delget, deledit, delta, or create, one will be supplied.\n"
    puts "  FILENAMES:"
    puts "   SCCS files to check in or out"
    puts "\n\n"
    exit
}

foreach arg $argv {
    set i [lsearch -exact $argv $arg]
    switch -glob -- $arg {
	-fixdev {set fixdev 1; incr numOpts}
	-fixpub {set fixpub 1; set user public; incr numOpts}
	-fixdist {set fixdist 1; set user dist; incr numOpts}
	-checkPub {set checkPub 1; set user public; incr numOpts}
	-checkDev {set checkDev 1; incr numOpts}
	-public {set makePublic 1; set user public; incr numOpts}
        -outside {set makeOutside 1; set user dist; incr numOpts}
	-update {set update 1; incr numOpts}
	-* {lappend sccsOpts $arg}
	info {set showInfo 1; incr numOpts}
	create {set sccsCmd "sccs create"; incr numOpts}
	get {set sccsCmd "sccs get"; incr numOpts}
	delget {set sccsCmd "sccs delget"; incr numOpts}
	delta {set sccsCmd "sccs delta"; incr numOpts}
	deledit {set sccsCmd "sccs deledit"; incr numOpts}
	unedit {set sccsCmd "sccs unedit"; incr numOpts}
	enter {set sccsCmd "sccs enter"; incr numOpts}
	edit {set sccsCmd "sccs edit"; incr numOpts}
	print {set sccsCmd "sccs print"; incr numOpts}
	clean {set sccsCmd "sccs clean"; incr numOpts}
        default {lappend tmpFileList $arg}
    }
}

# Check to see how many options were passed to websccs
if {$numOpts != 1} {
    puts "ERROR: Too many options passed to websccs"
    puts "You may only specify one of the following at a time:"
    puts "    -fixdev, -fixpub, -fixdist, -checkDev, -checkPub, -public, -update, command"
    puts "You specified $numOpts!"
    exit
}

# Set up variable values based on the current user
setVarValues $user
if {$debug} {
    puts "Variable Values (User = $user):"
    puts "--------------------------------"
    foreach var $fixVarList {
	puts "  $var = $VARS($var)"
    }
}

# Read files in distribution
getDist $distFile
if {$makeOutside == 1} {
    filterDist
}
setDistFileNames $user
if {$debug} {
    puts "\nDistribution:"
    puts "--------------"
    foreach file $distList {
	puts "  File: $file ($DIST($file,mode), $DIST($file,dir), $DIST($file,fullPath))"
    }
}

# Check for existence of -y option and add where necessary
switch -regexp $sccsCmd {
    (delget|deledit|create|delta) {
	if {[lsearch $sccsOpts -y*] == -1} {
	    # Add dummy -y argument
	    append sccsOpts "-y"
	}
    }
    default {}
}

# Verify that files specified in the SCCS file list exist
if {$tmpFileList != ""} {
    switch -regexp $sccsCmd {
	(deledit|delget|create|delta|enter) {
	    foreach file $tmpFileList {
		if [info exists DIST($file,fullPath)] {
		    set fileName $DIST($file,fullPath)
		} else {
		    set fileName [file join $workDir $file]
		    set DIST($file,fullPath) $fileName
		}

		if [file exists $fileName] {
		    lappend tmpSccsFileList $file
		} else {
		    puts "\nWARNING: SCCS file \"$file\" does not exist ... ignoring"
		}
	    }
	}
	default {set tmpSccsFileList $tmpFileList}
    }
}

# If creating new files, modify admin permissions to permit all dev users
#  to create deltas
if {$sccsCmd == "sccs create" || $sccsCmd == "sccs enter"} {
    foreach usr $devUsers {
	append sccsOpts " -a$usr"
    }
}

# Determine checked out sccs files
puts "\nCurrently checked out SCCS Files:"
puts "================================="
set sccsChkList ""
foreach usr $devUsers {
    set sccs($usr) [getSccsInfo $usr]
    if {$sccs($usr) != ""} {
	foreach file $sccs($usr) {
	    lappend sccsChkList $file
	}
	if {[llength $sccs($usr)] > 0} {
	    puts " $usr: $sccs($usr)"
	}
    }
}
if {$sccsChkList == ""} {
    puts "  None"
    set sccsFileList $tmpSccsFileList
} else {
    # Check command-line files to make sure they aren't checked out by anyone else
    #  This is only relevant for sccs commands that wish to modify these files.
    #  For 'sccs get', it is ok...
    foreach file $tmpSccsFileList {
	set checkedOut 0
	foreach usr $devUsers {
	    if {$usr != $user} {
		if {[lsearch -exact $sccs($usr) $file] != -1} {
		    set checkedOut 1
		}
	    }
	}

	if {!($checkedOut) || [string match *get* $sccsCmd]} {
	    lappend sccsFileList $file
	} else {
	    puts "\nFile \"$file\" is checked out by another user"
	}
    }
}


##################################################################
#         ACTUAL WORK GETS DONE BELOW THIS LINE!!!
##################################################################
# Exit if info requested
if {$showInfo} {
    exit
}

# Check permissions in public tree
if {$checkPub} {
    set user public
    setVarValues public
    checkDist public
    exit
}

# Check permissions in public tree
if {$checkDev} {
    checkDist $user
    exit
}

# Fix variables in source code
if {$fixdev || $fixpub || $fixdist} {
    puts "Updating variables in source code"
    puts "  User: $user   Tree: $VARS(wsRootDir)"
    puts "========================================================="
    fileUpdate $distList $fixVarList
    fixPerms $VARS(wsRootDir) $distList
    if {$user == "public"} {
	fixOwn $VARS(wsRootDir) $distList $serverUser $serverGroup
    }
    if {$fixpub} {
	fixSrc $DEV(public)
    }
    if {$fixdist} {
	fixSrc $DEV(dist)
    }
    if {$fixdev} {
	fixSrc $DEV($user)
    }
    exit
}


# Create clean public distribution (for NIST)
if {$makePublic} {
    # Check to make sure no SCCS files are checked out
    set pubList ""
    if {$sccsChkList != ""} {
	puts "\nWARNING: Only files not currently checked out will be copied..."
	if {$sccsFileList == ""} {
	    # Compare against full distribution
	    set srchList $distList
	} else {
	    # Compare against sccs file list
	    set srchList $sccsFileList
	}

	foreach file $srchList {
	    if {[lsearch -exact $sccsChkList $file] == -1} {
		lappend pubList $file
	    } else {
		puts "    $file checked out...not copied"
	    }
	}
    } else {
	if {$sccsFileList == "" } {
	    set pubList $distList
	} else {
	    set pubList $sccsFileList
	}
    }

    #################################
    #  PUBLIC RELEASE CODE (for NIST)
    #################################
    # Copy distribution files to public directories
    cd $sccsDir
    if [catch {sccsGetDist $publicDir $pubList} err] {
	puts "Error copying to public tree:\n$err"
	exit
    }

    # Change permissions on files
    fixPerms $publicDir $pubList

    # Fix paths
    fileUpdate $pubList $fixVarList

    # Update source commands
    if [catch {fixSrc $DEV(public)} err] {
	puts "Error updating source commands: $err"
    }

    # Fix file ownership to match HTTP server requirements
    fixOwn $publicDir $pubList $serverUser $serverGroup

    exit
}
	
# Create clean public distribution (for outside use)
if {$makeOutside} {
    # Check to make sure no SCCS files are checked out
    set pubList ""
    if {$sccsChkList != ""} {
	puts "\nWARNING: Only files not currently checked out will be copied..."
	if {$sccsFileList == ""} {
	    # Compare against full distribution
	    set srchList $distList
	} else {
	    # Compare against sccs file list
	    set srchList $sccsFileList
	}

	foreach file $srchList {
	    if {[lsearch -exact $sccsChkList $file] == -1} {
		lappend pubList $file
	    } else {
		puts "    $file checked out...not copied"
	    }
	}
    } else {
	if {$sccsFileList == "" } {
	    set pubList $distList
	} else {
	    set pubList $sccsFileList
	}
    }

    #######################
    #  OUTSIDE RELEASE CODE
    #######################
    # Copy distribution files to outside distribution directories

    cd $sccsDir
    if [catch {sccsGetDist $distDir $pubList} err] {
	puts "Error copying to release tree:\n$err"
	exit
    }
    
    # Change permissions on files
    fixPerms $distDir $pubList

    # Fix paths
    fileUpdate $pubList $fixVarList

    # Update source commands
    if [catch {fixSrc $DEV(dist)} err] {
	puts "Error updating source commands: $err"
    }

    # Fix file ownership to match HTTP server requirements
    fixOwn $distDir $pubList $serverUser $serverGroup

    exit
}

##################################
# UPDATE DEVELOPER DISTRIBUTION
##################################
if {$update} {
    # Check to make sure no SCCS files are checked out
    set devList ""
    if {$sccsChkList != ""} {
	puts "\nWARNING: Only files not currently checked out will be copied..."
	if {$sccsFileList == ""} {
	    # Compare against full distribution
	    set srchList $distList
	} else {
	    # Compare against sccs file list
	    set srchList $sccsFileList
	}

	foreach file $srchList {
	    if {[lsearch -exact $sccsChkList $file] == -1} {
		lappend devList $file
	    } else {
		puts "    $file checked out...not copied"
	    }
	}
    } else {
	if {$sccsFileList == "" } {
	    set devList $distList
	} else {
	    set devList $sccsFileList
	}
    }

    # Copy distribution files to developer directory
    cd $sccsDir
    if [catch {sccsGetDist $DEV($user) $devList} err] {
	puts "Error copying to developer tree:\n$err"
	exit
    }

    # Change permissions on files
    fixPerms $DEV($user) $devList

    # Fix paths
    fileUpdate $devList $fixVarList

    # Fix source commands
    if [catch {fixSrc $DEV($user)} err] {
	puts "  Error updating source commands: $err"
    }
    exit
}


######################################################
#  SCCS Actions on files
######################################################
# Make sure user has specified valid SCCS files
if {[llength $sccsFileList] == 0} {
    puts "\nNo valid SCCS files specified...exiting"
    exit
}

if {$sccsCmd != ""} {
    if {$sccsOpts != ""} {
	set sccsCmd [join [list $sccsCmd $sccsOpts]]
    }
} else {
    puts "\nNo SCCS command specified...exiting"
    exit
}

# Perform SCCS actions
puts "\nSCCS Command: $sccsCmd"
puts "SCCS file list: $sccsFileList"
puts "============================================="
cd $sccsDir
foreach file $sccsFileList {
    set devFile $DIST($file,fullPath)

    # Check to see if parent directory exists and create if not
    set parentDir [file dirname $devFile]
    if ![file exists $parentDir] {
	if ![catch {file mkdir $parentDir} err] {
	    file attributes $parentDir -permissions 000775
	    file attributes $parentDir -group websbmt
	} else {
	    puts "ERROR: $err"
	    continue
	}
    }

    if {$debug} {
	puts " Developer filename: $devFile"
    }

    # Execute SCCS command for $file
    if [regexp {^sccs (get|edit).*} $sccsCmd match] {
	catch {eval exec $sccsCmd $file} result
	if [string match *ERROR* $result] {
	    puts "\nSCCS Error ($file): \n $result\n"
	} else {
	    puts "$file\n$result\n"
	    file copy -force $file $devFile
	}
    } elseif [regexp {^sccs (create|enter|delta|delget|deledit).*} $sccsCmd match cmd] {
	# 1. Copy file from user --> sccsDir
	# 2. Delete user copy of file
	# 3. Execute SCCS command
	# 4. Copy file from sccsDir --> user
	catch {file copy -force $devFile $sccsDir
	       eval exec $sccsCmd $file} result
	if [string match *ERROR* $result] {
	    puts "\nSCCS Error ($file): $result\n"
	} else {
	    puts "$file \n$result\n"
	}

	switch -glob $cmd {
	    *delget* {file copy -force $file $devFile}
	    *deledit* {file copy -force $file $devFile}
	    *create* {file copy -force $file $devFile}
	    default {}
	}
    } else {
	catch {eval exec $sccsCmd $file} result
	if [string match *ERROR* $result] {
	    puts "\nSCCS Error ($file): \n $result\n"
	} else {
	    puts "$file \n$result\n"
	}
    }

    # Remove SCCS file from SCCS Directory
    catch {file delete -force $file}
}

# Fix up variable names and file permission
switch -glob $sccsCmd {
    *unedit* {}
    default {
	fixPerms $DEV($user) $sccsFileList
	fileUpdate $sccsFileList $fixVarList
	if [catch {fixSrc $DEV($user)} err] {
	    puts "Error updating source: $err"
	}
    }
}












