#!/bin/sh
#  Restart using local wish interpreter \
exec wish "$0" "$@"

#######################
# DIRECTORY SETUP
set installDir "/sasp/WebSubmit/develop/devryanm/install"
set wsRootDir "/sasp/WebSubmit/webSubmit2.5"
set wsLibDir "/sasp/WebSubmit/webSubmit2.5/lib"


#######################
# Initialization
#######################

# Variables
set debug 0

# Initialize wsManager using manager.setup and websubmit.paths
#  Contains variable definitions and auto_path setup for wsManager
set init ""
set pathFile [file join $installDir websubmit.paths]
set setupFile [file join $installDir manager.setup]
set defaultFile [file join $installDir websubmit.defaults]
set siteFile [file join $installDir websubmit.site]
foreach file [list $pathFile $setupFile] {
    if ![file exists $file] {
	puts "Setup file ($file) is missing or unreadable...exiting"
	exit
    } else {
	if [catch {source $file} err] {
	    puts "Setup file ($file) could not be sourced...exiting"
	    exit
	}
	append init "Loaded $file...\n"
    }
}

# Source the default configuration file (provided with the package)
if [file exists $defaultFile] {
    if [catch {source $defaultFile} err] {
	puts "Error sourcing defaults file ($defaultFile):\n$err"
    }
    append init "Loaded defaults: $defaultFile...\n"
}

# Site configuration file (possibly different from defaults)
set notConfigured 1
if [file exists $siteFile] {
    set notConfigured 0
    if [catch {source $siteFile} err] {
	puts "Error sourcing site configuration file:\n$err"
    }
    append init "Loaded site-specific configuration...\n"
}


########################################################
# Packages Required
#  auto_path should be configured correctly by 
#    websubmit.paths or config.defaults
########################################################
package require Widget
package require wsConfigure


#################################################################
# Perform some simple checks to see if the site's configuration
#  has been created, and if WebSubmit exists
#################################################################
# Check to see if WebSubmit has been configured for this site yet
#  If not, then database manipulation will be disabled, since paths
#  in WebSubmit probably do not point to the proper DB files
if {$notConfigured} {
    append init "\nWARNING: Site has not been configured yet...disabling database manager\n\
	    Update configuration using the Configuration Manager --> Update\n\n"
}

# Load WebSubmit package for database manipulation
set noWebSubmit 0
if [catch {verifyPackage webSubmit} err] {
    append init "\nWARNING: WebSubmit package could not be loaded: $err\n\
	    Database manager will be disabled until installed properly\n\n"
    set noWebSubmit 1
}
append init "Initialization complete"


###################################################################################
#                                GUI CONSTRUCTION
###################################################################################

# Load X resource definitions
option readfile [file join $installDir manager.xresources] 40
set xtitle "WebSubmit Manager"
wm title . $xtitle



#########################
# COLORS and DIMENSIONS  
#########################
set titleColor "darkslateblue"
set titleBgColor "linen"
set titleBd "2"
set primaryTabColor "ivory1"
set tabDisabledColor "dimgray"
set tier1Color "$titleBgColor"
set tier2Color "$titleBgColor"
set tier1Bd "2"
set tier2Bd "3"
set tabCoverColor "ivory3"
set secondaryTabColor "ivory3"
set sectionHdColor plum4
set outputFgColor "darkslateblue"
set outputErrColor "mediumslateblue"
set outputBgColor "grey"
set tabWidHeight 20
set entryWidth 65
set labelWidth 30

#####################
# Frame definitions #
#####################
frame .t -bg $titleBgColor -relief sunken -bd $titleBd
frame .m
frame .b -bd 2
set top .t
set mid .m
set bot .b
pack $top -side top -fill both -padx 1m
pack $mid -side top -fill both -expand 1 -padx 1m
pack $bot -side top -fill both -padx 1m -pady 3m

# Top Frame
label $top.label -text "WebSubmit Manager" -fg $titleColor -bg $titleBgColor -font {tekton 18 bold}
button $top.exit -text "Quit" -command "exit" 

# Middle Frame - Tab Notebook
set tab [tabnotebook $mid.tab -padx 10 -pady 5 -bd $tier1Bd\
	-activebackground $primaryTabColor -bg $tier1Color -disabledbackground $tabDisabledColor]
$tab.hold configure -background $tabCoverColor

# Bottom Frame - Text output window
set outputText $bot.text
scrolledtext $outputText -font {Courier 12 normal} -bg $outputBgColor -fg $outputFgColor -height 10 -autoscrollbar 0
$outputText.sx configure -background $outputBgColor -activebackground ivory3 -troughcolor azure3
$outputText.sy configure -background $outputBgColor -activebackground ivory3 -troughcolor azure3
$outputText.text tag configure error -foreground $outputErrColor -lmargin1 10 -lmargin2 15

# Main widget Packing
pack $top.label -side left -padx 2m -pady 1m
pack $top.exit -side right -padx 2m -pady 1m
pack $tab -fill both -expand 1
pack $bot.text -fill both -expand 1


#############################
#   CONFIGURATION MANAGER   # 
#############################
set configNote [addTab $tab t {Configuration Manager} notebook\
	-pady 5 -bg $tier2Color -activebackground $secondaryTabColor\
	-disabledbackground $tabDisabledColor -bd $tier2Bd]
set configMan $configNote.tab

# Tcl Setup
set configTcl [addTab $configMan tcl {Interpreters and Libraries} textMegawidget TCLSETUP -height $tabWidHeight]

# WebSubmit Setup
set configWs [addTab $configMan ws {WebSubmit Attributes} textMegawidget WSSETUP -height $tabWidHeight]

# UNIX Commands
set configUnix [addTab $configMan unix {UNIX Commands} textMegawidget UNIXSETUP -height $tabWidHeight]

# HTTP Server Info
set configHttp [addTab $configMan http {Web Servers} textMegawidget HTTPSETUP -height $tabWidHeight]

# Updater
set configUpdate [addTab $configMan update {Update} textMegawidget UPDATE -height $tabWidHeight]


####################################################################
# The following notebooks require WebSubmit for their functionality
#  If WebSubmit is not available, disable the tabs.  They will be
#  enabled after updateConfig is called (assuming success)
####################################################################
set dbNote [addTab $tab t2 {Database Manager} notebook\
	-pady 5 -bg $tier2Color -activebackground $secondaryTabColor\
	-disabledbackground $tabDisabledColor -bd $tier2Bd]
set dbMan $dbNote.tab

set packNote [addTab $tab t3 {Package Manager} notebook\
	-pady 5 -bg $tier2Color -activebackground $secondaryTabColor\
	-disabledbackground $tabDisabledColor -bd $tier2Bd]
set packMan $packNote.tab

###################################################################
#                      DATABASE MANAGER                           #
###################################################################

# User and Host List Manager
set userMan [addTab $dbMan userMan {Users and Hosts} dbManager]
set userText $userMan.db.text
set authBuild "authTabDb $userMan $userText"
set BUILDTABS(::webSubmit::auth::DB,parent) $userText
set BUILDTABS(::webSubmit::auth::DB,build) $authBuild
set DBMAP($userText,NS) ::webSubmit::auth
set DBMAP($userText,dbMake) ::webSubmit::auth::createAuthDb
set DBMAP($userText,dbKeyList) ::webSubmit::auth::dbKeyList
set DBMAP($userText,attrList) ::webSubmit::auth::attrList
set DBMAP($userText,fieldList) ::webSubmit::auth::fieldList
set DBMAP($userText,DB) ::webSubmit::auth::DB
set DBMAP($userText,made) 0
set DBMAP($userText,keyText) "User ID"
set DBMAP($userText,displayList) ::webSubmit::auth::hostList

# Master Database Manager
set masterMan [addTab $dbMan masterMan {Master Database} dbManager]
set masterText $masterMan.db.text
set masterBuild "tabDb $masterMan $masterText"
set BUILDTABS(::webSubmit::master::DB,parent) $masterText
set BUILDTABS(::webSubmit::master::DB,build) $masterBuild
set DBMAP($masterText,NS) ::webSubmit::master
set DBMAP($masterText,dbMake) ::webSubmit::master::createMasterDb
set DBMAP($masterText,dbKeyList) ::webSubmit::master::dbKeyList
set DBMAP($masterText,attrList) ::webSubmit::master::attrList
set DBMAP($masterText,fieldList) ::webSubmit::master::fieldList
set DBMAP($masterText,DB) ::webSubmit::master::DB
set DBMAP($masterText,made) 0
set DBMAP($masterText,keyText) "Record:"
set DBMAP($masterText,displayList) {}

# Activate DB tabs if possible
set dbTabList [list $userMan ::webSubmit::auth::DB $masterMan ::webSubmit::master::DB]
if {$notConfigured || $noWebSubmit} {
    foreach {dbTab dbName} $dbTabList {
	$dbMan itemconfigure $TABS($dbMan,$dbTab) -state disabled
    }
} else {
    # Activate the tabs
    activateDbTabs
}

#############################
#     PACKAGE MANAGER       # 
#############################
# Editor for queueing system packages
set fileInfoList [list "Load Leveler" wsLoadLeveler.tcl "NQS" wsNQS.tcl "LSF" wsLSF.tcl]
set queueEdit [addTab $packMan queue {Queuing Systems} tabEditor $fileInfoList wsLibDir]

# Editor for networking
set fileInfoList [list "Network" wsNetwork.tcl]
set networkEdit [addTab $packMan network {Networking} tabEditor $fileInfoList wsLibDir]

# Editor for general packages
set fileInfoList [list "WebSubmit" webSubmit.tcl "Utilities" wsUtil.tcl "Database" wsDatabase.tcl\
	"Authentication" wsAuth.tcl "CGI" wsCgi.tcl "Master Page" wsMaster.tcl "Forms" wsForms.tcl]
set genEdit [addTab $packMan general {General Packages} tabEditor $fileInfoList wsLibDir]



#####################
# Enter event loop
writeOutput $init









