# WebSubmit 2.0 wsAuthtcl @(#)wsAuth.tcl	1.26 SCCS/s.wsAuth.tcl  98/12/18  

# AUTHORS #############################################################

#  Authors: John Koontz, Ryan McCormack, Judith Devaney

# Copyright Information ###############################################
 
# This software was developed at the National Institute of Standards
# and Technology by employees of the Federal Government in the course
# of their official duties.  Pursuant to title 17 Section 105 of the
# United States Code this software is not subject to copyright
# protection and is in the public domain.

# CODEINFO BEGIN ######################################################

# info package wsAuth> This file contains a package to handle
#     processing of WebSubmit access authorizaions.

# CODEINFO END ########################################################

# Package Declarations ################################################

package provide wsAuth 1.0

# Namespace Definition ################################################
namespace eval webSubmit::auth {
    # Procedures to be exported
    namespace export createAuthDb
    namespace export authTrans
    namespace export isAdmin
    namespace export hasCert

    # Namespace variable definitions
    variable thisNS [namespace current]
    variable upNS [namespace parent]
    variable wsNS [namespace parent]

    # Procedures to be imported
    namespace import ${wsNS}::database::*
    namespace import ${wsNS}::util::*

    # Variable aliasing
    alias $wsNS wsLibDir wsID userName userType userPath userEmail wsUserDir adminMailAddr wsCertDir certManager

    # General variable definitions
    variable userName
    variable userEmail
    variable userType
    variable debug 0
    variable hostList
    readOnly issuerIdHead "issuer"
    readOnly adminIdHead "admin"
    ;;readOnly certEnvName "CLIENT_CERT"
    variable dump 1

    # Database variables
    variable DB
    variable DBTXT
    variable dbFile "$wsLibDir/auth"
    variable txtFile ""
    variable stateFile ""
    variable reqAttrList [list userType userDir userName email status]
    variable attrList
    variable dbKeyList
}

# Procedure definitions ###############################################

# CODEINFO BEGIN ######################################################

# info proc createAuthDb> Procedure to parse and store database related 
#     to user authentication and access authorization

# return> Status

# loc-sclr thisNS> Name of authentication namespace
# authNmspace-arr DB> Array of user information
# authNmspace-sclr attrList> List of actual attributes for auth DB
# authNmspace-sclr reqAttrList> List of required attributes for auth DB
# authNmspace-list dbKeyList> List of websubmit user IDs

proc webSubmit::auth::createAuthDb {} {
    variable thisNS 
    variable wsNS

    # Variable declarations
    foreach var [list dbKeyList attrList DB debug] {
	variable $var
    }

    # Read Database Information
    buildDb  [file join [set ${wsNS}::auth::wsLibDir] auth] $thisNS

    # Create list of valid hosts
    getHosts

    return 0
}



# CODEINFO BEGIN ######################################################

# info proc authTrans> Procedure to translate a valid websubmit user ID 
#     into a login on a specified remote hostname

# arg hostName> The name of the remote host
# return> Login name for current websubmit user on hostName OR Error

proc webSubmit::auth::authTrans {hostName} {
    variable DB
    variable wsID
    variable attrList

    # Verify that host is in auth DB
    if {[lsearch -exact $attrList $hostName] == -1} {
	return -code error "ERROR(authTrans): \"$hostName\" is not a valid remote host"
    }

    if [info exists DB($wsID,$hostName)] {
	return $DB($wsID,$hostName)
    } else {
	return -code error "ERROR(authTrans): Data unavailable for ID=\"$wsID\" host=\"$hostName\""
    }
}


# CODEINFO BEGIN ######################################################

# info proc getHosts> Procedure to return list of WebSubmit target hosts

# return> List of WS hosts in auth.db

proc webSubmit::auth::getHosts {} {
    variable attrList 
    variable reqAttrList
    variable hostList
    
    catch {unset hostList}
    foreach var $attrList {
	if {[lsearch -exact $reqAttrList $var] == -1} {
	    # Host found
	    lappend hostList $var
	}
    }
    return 0
}


# CODEINFO BEGIN ######################################################

# info proc isAdmin> Procedure to determine if a user has a valid
#     administrator wsID

# arg> None
# return> 1 = admin user, 0 = regular user

proc webSubmit::auth::isAdmin {} {
    variable wsID
    variable adminIdHead
    if [string match ${adminIdHead}* $wsID] {
	return 1
    } else {
	return 0
    }
}


# CODEINFO BEGIN #####################################################
# info proc authenticate> Procedure to retrieve and verify a user's
#     WebSubmit identification number

# arg> None
# return> Success = 1, Failure = Bomb

proc webSubmit::auth::authenticate {} {
    variable dump
    foreach var [list wsID userName userType userPath userEmail adminMailAddr\
	    certManager wsCertDir DB dbKeyList thisNS wsNS wsUserDir debug] {
	variable $var
    }

    # Verify HTTPS connection
    if ![string match {[Oo][Nn]} [getEnv HTTPS]] {
	# HTTPS off
	cgiBomb authenticate "Insecure connection (accessed via HTTP) - Exiting"
    }

    # Get certificate data
    set certData [getClientCert]

    # Hash certificate
    set certHash [hash $certData]

    # Create authentication database
    createAuthDb

    # Look for certificate hash in authentication database
    if {[lsearch -exact $dbKeyList $certHash] == -1} {
	if {$dump} {
	    # Dump certificate to disk for viewing by admin
	    set certDumpName "cert.[pid]"
	    certDump $certData $certHash $certDumpName
	    
	    # Mail the WebSubmit administrator
	    set certFile [file join $wsCertDir $certDumpName]
	    cgi_mail_start $adminMailAddr
	    cgi_mail_add "Subject: Unregistered WebSubmit user"
	    cgi_mail_add
	    cgi_mail_add "An unregistered user has attempted to access the system."
	    cgi_mail_add "The user's certificate has been dumped to\n\n $certFile\n"
	    cgi_mail_add "Check the user's data and add them to the system if desired."
	    cgi_mail_end
	}
    } else {
	# Make sure user's status is active
	if {[getDbVal $thisNS $certHash status] != "active"} {
	    cgiBomb authenticate "The WebSubmit server indicates your status as inactive.\
		    Please contact the WebSubmit administrator if you feel this is incorrect."
	} else {
	    # Set up immutable data related to user
	    readOnly wsID $certHash
	    readOnly userName [getDbVal $thisNS $wsID userName]
	    readOnly userPath [file join $wsUserDir [getDbVal $thisNS $wsID userDir]] $wsNS
	    readOnly userEmail [getDbVal $thisNS $wsID email] $wsNS
	    readOnly userType [getDbVal $thisNS $wsID userType] $wsNS
	}
    }

    if {$debug} {
	puts "\nAuthentication:"
	puts "---------------"
	foreach var [list wsID userName userPath userEmail userType] {
	    puts "  $var: [set $var]"
	}
    }
    return 1
}

# CODEINFO BEGIN ######################################################
# info proc getClientCert> Procedure to retrieve a client's certificate	    
# arg> None
# return> Client digital certificate or error

proc webSubmit::auth::getClientCert {} {
    variable certEnvName

    # Check for existence of Certificate
    set certData [getEnv $certEnvName]
    if {$certData != ""} {
	regsub -all "\n" $certData "" certData
    }
    return $certData
}


# CODEINFO BEGIN ########################################################
# info proc certDump> Procedure to take a base-64 encoded certificate,
#     decode it, and dump the data to disk for later use.  This is 
#     primarily intended as an easy way to notify the WebSubmit
#     administrator if a new user needs to be registered with the system.

# arg base64cert> A base-64 encoded version of the certificate
# arg hash> The MD5 hash of this certificate
# arg dumpFile> File in $wsCertDir in which data is dumped
# return> None

proc webSubmit::auth::certDump {base64cert hash dumpFile} {
    variable wsCertDir

    # Dump cert to temp file for reading by certutil
    set tmpFile [file join $wsCertDir tmp.[pid]]
    if [catch {open $tmpFile w} fd] {
	return -code error "Could not open temp file for cert"
    } 
    puts $fd $base64cert
    close $fd

    # Decode certificate
    set certExec [file join $wsCertDir certutil]
    if [catch {exec $certExec $tmpFile 2>@ stdout} allData] {
	return -code error "Could not decode certificate $tmpFile"
    }

    # Set up fields to retrieve from certificate
    set CERT(fieldList) [list "issuer dn" "subject dn"]

    # Parse cert for specified fields
    set fieldData [split $allData \n]
    foreach field $CERT(fieldList) {
	set fieldIdx [lsearch -regexp $fieldData "^\[\t \]*$field *:.*"]
	if {$fieldIdx == -1} {
	    set CERT($field) "Unknown"
	    continue
	} else {
	    if [regexp "$field *: *(.*)" [lindex $fieldData $fieldIdx] match data] {
		puts "Field Data: $data"
		set CERT($field) $data
	    } else {
		set CERT($field) "Unknown"
	    }
	}
    }
    
    # Dump certificate data to tmpFile
    set certFile [file join $wsCertDir $dumpFile]
    if [catch {open $certFile w} fd] {
	return -code error "could not open dump file for certificate"
    }
    puts $fd "Certificate Data:"
    puts $fd "-----------------"
    foreach field $CERT(fieldList) {
	if {$field == "issuer dn" || $field == "subject dn"} {
	    set fList [split $CERT($field) ,]
	    foreach datum $fList {
		if {$field == "subject dn"} {
		    # Parse CN string for special cases
		    if [regexp { *CN *= *(.*)/Email=(.*)} $datum match name email] {
			set tmp [string trim $name]
			if [regexp {([a-zA-Z0-9_]+) *\((.*)\)} $tmp match dir name] {
			    set CERTDATA(name) [string trim $name]
			    set CERTDATA(dir) $dir
			} else {
			    set CERTDATA(name) $tmp
			    set CERTDATA(dir) "*"
			}
			set CERTDATA(email) [string trim $email]
		    } elseif [regexp { *CN *=(.*)} $datum match cn] {
			set CERTDATA(name) [string trim $cn]
			set CERTDATA(email) "*"
			set CERTDATA(dir) "*"
		    }
		}
	    }
	}
    }
    puts $fd "$allData\n\n"
    
    set header [format "%-32s %1s %-5s %1s %-3s %1s %-20s %1s %-20s %1s %-6s" "CERTIFICATE DIGEST"\
	    : "TYPE" : "DIR" : "SUBJECT NAME" : "SUBJECT EMAIL" : "STATUS"]
    puts $fd "$header\n"
    set data [format "%32s %1s %-3s %1s %-8s %1s %-20s %1s %-20s %1s %-6s" $hash :\
	    user : $CERTDATA(dir) : $CERTDATA(name) : $CERTDATA(email) : active]
    puts $fd "$data"

    # Cleanup
    close $fd
    file delete $tmpFile

    return
}

# CODEINFO BEGIN #####################################################
# info proc hasCert> Procedure to check whether a certificate exists
#     in the environment

proc webSubmit::auth::hasCert {} {
    variable certEnvName
    global env
    if [info exists env($certEnvName)] {
	return 1
    } else {
	return 0
    }
}


