# WebSubmit 2.0 wsDatabase.tcl @(#)wsDatabase.tcl	1.43 SCCS/s.wsDatabase.tcl  99/07/12 

# AUTHORS #############################################################

#  Authors: John Koontz, Ryan McCormack, Judith Devaney
#  Original Development Work: Robert Lipman, John Koontz, Katherine
#      Pagoaga

# Copyright Information ###############################################
 
# This software was developed at the National Institute of Standards
# and Technology by employees of the Federal Government in the course
# of their official duties.  Pursuant to title 17 Section 105 of the
# United States Code this software is not subject to copyright
# protection and is in the public domain.

# CODEINFO BEGIN ######################################################

# info package wsDatabase> This file contains a package of database 
#     routines to be used with the webSubmit package.  These routines
#     are used to:
#
#     - read databases, 
#     - read metadatabases (several related databases - called codatabases - 
#       combined together),
#     - arrange database information in the appropriate datastructures,  
#     - write databases back out, especially in cases where these
#       databases result from disassembling a metadatabase,
#     - serialize databases for quick rereading.
#     - dump various database data structures for debugging purposes,
#     - insert and retrieve individual database values, and test for
#       their existence,
#     - perform various other auxiliary tasks.
#
#     We use the term database here in the sense of a single table.
#     Tables consist of records, made up of fields.  Indexing fields 
#     are called keys.  Data fields are called attributes.
#
#     Databases are stored external to the program in files in 
#     metadatabase format, database format, and serialized format.  
#     Internally they are stored as database objects in database
#     namespaces.  External databases are identified by pathnames;
#     internal databases, by namespace names.  When an internal
#     database is set up, both names must be known.  Typically
#     the name of the namespace is the {tail {rootname pathName}},
#     i.e., the filename sans the extension.  Once the namespace
#     has been created, the namespace name alone suffices for 
#     internal use.  The pathName is stored in the namespace. 
#
#     The process of inputting database is:  
#
#     Routine                 Arguments      Computed   From NS
#
#     buildDb                 (NS) pathName
#       readMDb               NS                        dbFile
#         openDb              NS (pathName)  (pathName)         
#         configureDb         NS fieldList  
#         processCorecord     NS record ...
#           processRecord     NS record ...
#       readDb                NS                        dbFile
#         openDb              NS (pathName)  (pathName)
#         configureDb         NS fieldList
#         processRecord       NS record ...
#       readTxt               NS                        txtFile (deprecated)
#     writeDb                 NS (pathName)  (pathName)
#     serializeDb             NS (pathName)

# CODEINFO END ########################################################

# Package Declarations ################################################

package provide wsDatabase 1.0

# Namespace Definition ################################################
namespace eval webSubmit::database {
    namespace export buildDb
    namespace export configureDb
    namespace export dumpDbState
    namespace export dumpDbVal
    namespace export existDbVal
    namespace export getDbVal
    namespace export joinKey
    namespace export openDb
    namespace export processRecord
    namespace export readDb
    namespace export readMdb
    namespace export readTxt
    namespace export serializeDb
    namespace export setDbVal
    namespace export splitKey
    namespace export writeDb
    
    # General variable definitions
    variable thisNS [namespace current]
    variable upNS [namespace parent]
    variable wsNS [namespace parent]

    # Namespace imports
    catch {namespace import ${wsNS}::util::*}

    # Variable definitions
    variable debug 0
    variable stdConcatenator "_"
    variable paramConcatenator "."
    variable internalConcatenator "="
}

# Procedure definitions ###############################################

# CODEINFO BEGIN ######################################################

# info proc readDbMetainfo>  Extract a database metainformation line
#     from a database file. 

# arg dbFile>  Path to the database file.
# arg keyword>  The keyword used to pick the metainformation line out 
#    of database file dbFile
# return> Status

proc webSubmit::database::readDbMetainfo {dbFile keyword} {
    variable debug

    # Extract metainformation line
    if [catch {tclGrep "^$keyword" $dbFile} grepVal] {
	# No such metainformation line
	return {}
    } else {
	# Split metainformation into list without white-space
	set grepVal [compSplit [string trim $grepVal \{\}]]

	# Strip off keyword item and return rest of list
	return [lrange $grepVal 1 [llength $grepVal]]
    }
}


# CODEINFO BEGIN ######################################################

# info buildDb> Procedure to build a generic database.  Can handle a
#     build from a metadatabase (*.mdb), a regular database (*.db), or
#     a serialized database (*.tcl).  

# arg pathName>  Path to the database file.
# arg dbNS> Namespace name for database.  By default this is the rootdir
#     of the pathName, and located in the webSubmit namespace. 
# arg debugBuild> If true, trace the details of this build.
# return>  Final value of dbNS.

proc webSubmit::database::buildDb {pathName {dbNS ""} {debugBuild 0}} {
    variable debug
    variable wsNS
    variable stdConcatenator
    variable paramConcatenator

    # The namespace name dbNS supplied or created is assigned to the main
    # (type main) codatabase of a metadatabase, and to the only database
    # of simple database.  Auxiliary codatabases of a metadatabase 
    # receive their keyword as their name.  

    # If no namespace name dbNS provided, compute the namespace name from the 
    # pathname of the source file.
    if {"$dbNS" == ""} {
	set dbNS ${wsNS}::[nameDbNS $pathName]
    }

    # If the namespace name is relative, make an absolute namespace pathname
    # dbNS.  The default is to place the namespace in the calling environment.
    if {![string match "::*" "$dbNS"]} {
        set callingNS [uplevel 1 namespace current]
        set dbNS ${callingNS}::${dbNS}
        regsub "::::" $dbNS "::" dbNS
    }

    if {$debug} {
	puts "\nTRACE(buildDb): Database file is $pathName"
        if {$debugBuild} {
	    puts "TRACE(buildDb): Trace details of build of $dbNS."
	}
    }	

    # Determine which kind of database file (*.mdb. *.db, *.tcl) is being 
    # passed in.
    set extension [string range [file extension $pathName] 1 end]
    set pathName [file rootname $pathName]

    if {$debug} {
	puts "  Namespace = $dbNS"
	puts "  Extension = $extension"
	puts "  Stripped path = $pathName"
    }

    # Process the various cases determined by the file extension
    switch -exact $extension {  
	mdb {	
	    if {![fileUsable $pathName.mdb]} {
		set errMess "ERROR(buildDB):  File $pathName.mdb either\
                    missing or unreadable."
		return -code error $errMess
	    } else {
		readMdb $pathName $dbNS $debugBuild
	    }

	    alias $dbNS auxDbList

	    # Write the codatabase files into the directory from which
	    # the metadatabase file was supplied.
	    writeDb $dbNS $stdConcatenator
	    foreach auxDbNS $auxDbList {
		writeDb [namespace parent $dbNS]::${auxDbNS} $paramConcatenator
	    }

	    # Serialize the databases into the directory from which the 
	    # metadatabase was supplied.
	    serializeDb $dbNS
	    foreach auxDbNS $auxDbList {
		serializeDb [namespace parent $dbNS]::${auxDbNS}
	    }
	}

	db {
	    if {![fileUsable $pathName.db]} {
		set errMess "ERROR(buildDB):  File $pathName.db either\
                    missing or unreadable."
		return -code error $errMess
	    } else {
		readDb $pathName $dbNS
	    }

	    alias $dbNS auxDbList

	    # Read the auxiliary codatabases
 	    if [info exists auxDbList] {
		foreach auxDbNS $auxDbList {
                    set auxDb [file join [file dirnam $pathName] ${auxDbNS}]
		    if {![fileUsable $auxDb.db]} {
			set errMess "ERROR(buildDB):  File $auxDb.db, required\
                            with $pathName.db is either missing or unreadable."
			return -code error $errMess
		    } else {
			readDb $auxDb [namespace parent $dbNS]::${auxDbNS}
		    }			
		}
	    }

	    # Legacy processing of associated *.txt file (for master.db)
	    set txtFile $pathName.txt
	    if {![fileUsable $txtFile]} {
		set ${dbNS}::txtFile ""	
	    } else {
		set ${dbNS}::txtFile $txtFile
		readTxt $dbNS
	    }

	    # Serialize the database into the directory from which the 
	    # unserialized database was supplied.
	    serializeDb $dbNS
	    foreach auxDbNS $auxDbList {
		serializeDb [namespace parent $dbNS]::${auxDbNS}
	    }
	}

	tcl {
	    if {![fileUsable $pathName.tcl]} {
		set errMess "ERROR(buildDB):  File $pathname.tcl either\
                    missing or unreadable."
		return -code error $errMess
	    } else {
		source $pathName.tcl

		alias $dbNS dbFile auxDbList

		# Source the auxiliary codatabases
		if [info exists auxDbList] {
		    foreach auxDbNS $auxDbList {
			set auxDb [file join [file dirname $pathName]\
                            ${auxDbNS}.tcl]
			if {![fileUsable $auxDb]} {
			    set errMess "ERROR(buildDB):  File $auxDb,\
                                required with $pathName.tcl, is either\
                                missing or unreadable."
			    return -code error $errMess
			} else {
			    source $auxDb
			}
		    }
		}
	    }
	}
	
	"" {
	    if {[fileUsable $pathName.mdb]\
		&& ![fileCurrent $pathName.mdb $pathName.tcl]} {
		buildDb $pathName.mdb $dbNS
	    } elseif {[fileUsable $pathName.db]\
		&& ![fileCurrent $pathName.db $pathName.tcl]} {
		buildDb $pathName.db $dbNS
	    } elseif {[fileUsable $pathName.tcl]} {
		buildDb $pathName.tcl $dbNS
	    } else {
		set errMess "ERROR(buildDB): No database file $pathName.*"
		return -code error $errMess
	    }
	}

	default {
	    set errMess "ERROR(buildDB): No procedure available for building\
            databases from files with the extension \"$extension\"."
	    return -code error $errMess
	}
    }

    return $dbNS
}


# CODEINFO BEGIN ######################################################

# info readMdb> Procedure to read a generic metadatabase file 
#     containing metadatabase and main codatabase record attributes
#     lines, comments, and a set main and auxiliary codatabase records.  
#     The pathname root to the file holding the metadatabase is
#     supplied.  An extension of .mdb is assumed.

# arg pathName>  The root of the name of the metadatabase file.
# arg mainDbNS>  The name of the namespace associated with the main
#     codatabase.  
# arg debugBuild>  If true, trace details of this readMdb.
# return>  Status

proc webSubmit::database::readMdb {pathName mainDbNS debugBuild} {
    variable debug

    # Check for existence of metadatabase file
    if {![fileUsable $pathName.mdb]} {
	set errorMsg "ERROR(readMdb): File $pathName.mdb does not\
            exist or is not readable."
	return -code error $errorMsg              
    }

    # Extract metadatabase field descriptor into dbDescList
    set dbDescList [readDbMetainfo $pathName.mdb "::METADATABASE_ATTRIBUTES::"]
    if {[llength $dbDescList] == 0} {
	# Empty (or missing) field descriptor
	set errorMsg "ERROR(readMdb): Empty or missing metadatabase field\
            descriptor ($keyword line) in file $pathName.mdb."
	return -code error $errorMsg
    }

    # Build array of database descriptors from extracted list of descriptors
    # Form of descriptors:  

    # keyword : class (: storageMode)

    # The keyword is the word that precedes records of that type in the
    # metadatabase.  Except for the main codatabase (of class main),
    # this is also the name of the codatabase file. 

    # The class of the database indicates how to index it.
    # The available classes are:  

    # main - arbitrary key(s); arbitrary attributes
    #    Define the main database.
    # param - key $mainKey,$name; one attribute, value
    #    Define fixed or variable attributes of main codatabase records.
    # const - key $name; one attribute, value
    #    Define constants associated with the main database.

    # The storageMode of a codatabase record indicates whether successive
    # records with the same key repace each other (scalar) or append
    # (list).  Default is scalar.

    set descIndex 0 
    set auxDbList {}
    set keywordList {}
    set previousMain 0
    foreach dbDesc $dbDescList {
        set fieldList [trimSplit $dbDesc ":"]

        # Set defaults
        switch -exact [llength $fieldList] { 
	    1 {
		set errMess "ERROR(readMdb): Malformed METADATABASE_ATTRIBUTES\
                    line; no class specifier for $fieldList"
		return -code error $errMess
	    }
	    2 {
		lappend fieldList scalar
	    }
	    3 {
	    }	    
	    default {
		set errMess "ERROR(readMdb): Malformed METADATABASE_ATTRIBUTES\
                    line; more than three specifiers for $fieldList."
		return -code error $errMess
	    }
	}
	set keyword [lindex $fieldList 0]
        lappend keywordList $keyword
	if {"[lindex $fieldList 1]" == "main"  \
            && $previousMain} {
	    set errMess "ERROR(readMdb): There can only be one codatabase\
                using the class \"main\"."
	    return -code error $errMess
	}
	if {"[lindex $fieldList 1]" == "main"} {
	    set DESC($keyword,file) [namespace tail $mainDbNS]
	    set previousMain 1
	} else {
	    set theNS [lindex $fieldList 0]
	    set DESC($keyword,file) $theNS
	    lappend auxDbList $theNS
	}
	set DESC($keyword,class) [lindex $fieldList 1]
	set DESC($keyword,storageMode) [lindex $fieldList 2]

    }

    if {$debug} {
        dumpDesc $pathName $keywordList DESC 
    }

    # Open and configure the codatabases
    foreach keyword $keywordList {
	switch $DESC($keyword,class) {
	    main {
		# Open the namespace for the main codatabase.
		if {![nsExist $mainDbNS]} {
		    # No namespace at all
		    openDb $mainDbNS $pathName $debugBuild
		} elseif {![info exist ${mainDbNS}::dbFile]} {
		    # Namespace exists, but doesn't define dbFile, so update
		    openDb $mainDbNS $pathName $debugBuild
		}

		# Configure the main codatabase
		set fieldList \
                    [readDbMetainfo $pathName.mdb "::(RECORD|DB)_ATTRIBUTES::"]
		if {[llength $fieldList] == 0} {
		    # Empty (or missing) field descriptor
		    set errorMsg "ERROR(readMdb): Empty or missing database\
                        field descriptor ($keyword line) in file\
                        $pathName.mdb."
		    return -code error $errorMsg
		}

		configureDb $mainDbNS main $fieldList $auxDbList
	    }

	    param {
	        # Open the auxiliary codatabase namespaces and configure them
		set theNS $DESC($keyword,file)
		set theAbsNS [namespace parent $mainDbNS]::${theNS}
		if {![nsExist $theAbsNS]}  {
		    openDb ${theAbsNS} \
                        [file join [file dirname $pathName] $theNS] $debugBuild
		    set fieldList "key:elementName key:paramName value"
		}
		configureDb $theAbsNS param $fieldList
	    }

	    const {
		set theNS $DESC($keyword,file)
		set theAbsNS [namespace parent $mainDbNS]::${theNS}
		if {![nsExist $theAbsNS]}  {
		    openDb ${theAbsNS} \
                        [file join [file dirname $pathName] $theNS] $debugBuild
		    set fieldList "key:constName value"
		}
		configureDb $theAbsNS const $fieldList
	    }
	}
    } 

    # Open file and process data
    set mdbid [open "$pathName.mdb" r]
    set recordCount -1
    set corecord ""
    set scopeRecordKey "initial"
    while {[gets $mdbid line] >= 0} {
	# Trim whitespace from either end of line
	set line [string trim $line]

	# Parse lines of file
	switch -regexp $line {

	    {^(\#.*|::.*| *)$} {
		# Comment line, blank line, or :: line - ignore
	    } 

	    {^[-_.a-zA-Z0-9]+[ \t]+} {
		# New database corecord line:  alphabetic characters followed 
		# by a space followed by the corecord
		if {0 != [catch \
		    {regexp "^(\[-_.a-zA-Z0-9\]+)\[ \t\]+(.+)\$" $line match \
                        keyWord line} mess]} {
		    set errMess "ERROR(readMdb): Improperly formatted\
                        corecord line in $pathName.mdb:\
			\n    $mess"
		    return -code error $errMess
		}
		# If there's an existing corecord accumulated, push it out
		if {[string length $corecord] > 0} {
		    if {[llength [array names DESC $coKeyword,class]] == 0} {
			set errMess "ERROR(readMdb): Keyword $coKeyword undefined."
			return -code error $errMess
		    } else {
			if {"$DESC($coKeyword,class)" == "main"} {
			    set recordCount [incr recordCount]
			}
			set theNSTail $DESC($coKeyword,file)
			set theNS ::webSubmit::${theNSTail}
			set currentKey [processCorecord $theNS $corecord\
		            $DESC($coKeyword,class)\
                            $DESC($coKeyword,storageMode)\
                            $scopeRecordKey $recordCount]
			if {"$DESC($coKeyword,class)" == "main"} {
			    set scopeRecordKey $currentKey
			}
		    }
		}
		set coKeyword $keyWord
		set corecord [string trim $line]
	    }

	    {^\|[^\|]} {
		# Continuation line:  a '|' and then text
		if {[string length $corecord] == 0} {
		    set errMess "ERROR(readMdb): Unexpected corecord\
                        continuation line:\n    $line"
		    return -code error $errMess
		}
		set corecord [join [list [string trimright $corecord]\
                    [string trimleft [string range $line 1 end]]]]
	    }
	}
    }
    close $mdbid

    # If there's a last corecord accumulated, push it out
    if {[string length $corecord] > 0} {
	if {[llength [array names DESC $coKeyword,class]] == 0} {
	    set errMess "ERROR(readMdb): Keyword $coKeyword undefined."
	    return -code error $errMess
	} else {
	    if {"$DESC($coKeyword,class)" == "main"} {
		set recordCount [incr recordCount]
	    }
	    set theNSTail $DESC($coKeyword,file)
	    set theNS ::webSubmit::${theNSTail}
	    set curentKey [processCorecord $theNS $corecord\
		$DESC($coKeyword,class)\
                $DESC($coKeyword,storageMode)\
		$scopeRecordKey $recordCount]
	    if {"$DESC($coKeyword,class)" == "main"} {
		set scopeRecordKey $currentKey
	    }
	}
    }   

    # Output of database information
    if {$debug && [set ${mainDbNS}::debug]} {
	dumpDbVal $mainDbNS
    }
    foreach theNS $auxDbList {
	set auxDbNS [namespace parent $mainDbNS]::${theNS}
	if {$debug && [set ${auxDbNS}::debug]} {
	    dumpDbVal $auxDbNS
	}
    }

    return [expr [llength $auxDbList] + 1]
}


# CODEINFO BEGIN #######################################################

# info dumpDesc>  Print current state of DESC array.  

# arg dbFile>  Name of the metadatabase file.
# arg keywordList>  List of descriptor keywords 
# arg ptrDESC>  Pointer to the metadatabase descriptor array.
# return> None.
 
proc webSubmit::database::dumpDesc {dbFile keywordList ptrDESC} {
    upvar $ptrDESC DESC
       
    puts "\n"
    puts [header [list "Metadatabase Descriptors for File $dbFile.mdb"] \# -]
    if {[llength $keywordList] == 0} {
	puts "    No entries in the metadatabase descriptor array."
    } else {
	foreach dbDesc $keywordList {
	    puts "Descriptor for entry $dbDesc\
                 \n\----------------------------"
            puts "File name:      $DESC($dbDesc,file)"
	    puts "Class of file:  $DESC($dbDesc,class)"
	    puts "Storage mode:   $DESC($dbDesc,storageMode)"
	    puts "\n"
	}
    }
    return 0
}


# CODEINFO BEGIN #######################################################

# info processCorecord>  Process a metadatabase corecord, which is either
#     the main database corecord or an auxiliary corecord for the main
#     database or some other database, as specified in the metadatabase.

# arg NS>  Name of the corecord database namespace.
# arg corecord>  Text of corecord.
# arg class>  Class of corecord (main, param, const).
# arg storageMode>  Storage made (scalar, list).
# arg recordKey>  Key of the current main database record.
# arg recordCount>  Index of current main database record (0, 1, ...).
# return>  Key of processed record.

# local-sclr theRecord>  Constructed record passed to process*.
 
proc webSubmit::database::processCorecord {NS corecord class\
    storageMode recordKey recordCount} {
    variable debug
    variable stdConcatenator
    variable paramConcatenator

    # Perform the proper record processing for each type of corecord. 
    switch -exact $class {
	main {
	    return [processRecord $NS $corecord scalar $stdConcatenator\
			$recordCount]
	}

	param {
	    return [processRecord $NS "$recordKey : $corecord"\
                    $storageMode $paramConcatenator]
	}

	const { 
	    return [processRecord $NS $corecord $storageMode]
	}

	default {
	    set errMess "ERROR(processCorecord): Unrecognized corecord class\
                $class."
	    return -code error $errMess
	}
    }
}


# CODEINFO BEGIN #######################################################

# info readDb>  Read a simple general database file containing a record 
#     attributes line, comments, and a set of records.  The material 
#     read is inserted in a database namespace.  The root of the name 
#     of the database file is supplied.  An extension of .db is assumed.   
#
# 1)  Create a mapping from the concatenated key fields of a record
#     to the concatenated attribute fields.
#
# 2)  Augment the mapping to produce mappings from the concatenated
#     keys comma the attribute field name to the attribute field value for
#     each attribute in each record.
#
#  -  If there is a key field named auto, it is not specified in the
#     record, but an automatically generated index number (0, 1, 2, ...)
#     should be supplied for it.
#
#  -  If there is non-key field called auto, it, too, is not specified in
#     the record, but an automatically generated index number (0, 1, 2, ...)
#     should be supplied for it.
#
# If an input field has the value *, this is converted to "".  

# arg pathName>  The root of the name of the database file.
# arg dbNS>  Name of the namespace in which the database will be stored.
#     The file to be read should have been stored in $dbNS::dbfile by openDb
#     (or an ad hoc opening process).
# return> Status (Error = 1)

# loc-list fieldList> List of field names in database
# loc-sclr dbid> Filehandle for database file

proc webSubmit::database::readDb {pathName dbNS} {
    variable thisNS
    variable debug
    variable stdConcatenator
    variable paramConcatenator

    # Open the namespace for the database.
    if {![nsExist $dbNS]} {
	# No namespace at all
	openDb $dbNS $pathName
    } elseif {![info exist ${dbNS}::dbFile]} {
	# Namespace exists, but doesn't define dbFile, so update
	openDb $dbNS $pathName
    }

    # Check for existence of database file
    if {![fileUsable $pathName.db]} {
	set errorMsg "ERROR(readDb): File $pathName.db does not\
            exist or is not readable"
	return -code error $errorMsg              
    }

    # Extract database field descriptor
    set fieldList [readDbMetainfo $pathName.db "::(RECORD|DB)_ATTRIBUTES::"]
    if {[llength $fieldList] == 0} {
	# Empty (or missing) field descriptor
	set errorMsg "ERROR(readDb): Empty or missing database field\
            descriptor ($keyword line) in file $pathName.db."
	return -code error $errorMsg
    }

    # Extract list of auxiliary databases
    set auxDbList [readDbMetainfo $pathName.db "::AUXILIARY_DATABASES::"]

    # Extract database class
    set dbClassList [readDbMetainfo $pathName.db "::DATABASE_CLASS::"]
    if {[llength $dbClassList] == 0} {
	set dbClass main
    } else {
	set dbClass [lindex $dbClassList 0]
    }

    # Initialize the database namespace
    configureDb $dbNS $dbClass $fieldList $auxDbList

#     # Verify that database possesses required attributes
#     if ![reqList $attrList $reqAttrList] {
# 	# Database does not have proper attributes
# 	set errorMsg "ERROR(readDb): Database attributes ($attrList)\
#             do not match those required ($reqAttrList)"
# 	return -code error $errorMsg
#     }

    # Open file and process data
    if {"$dbClass" == "param"} {
	set keyConcatenator $paramConcatenator
    } else {
	set keyConcatenator $stdConcatenator
    }
    set dbid [open "$pathName.db" r]
    set index -1
    set record ""
    while {[gets $dbid line] >= 0} {
	# Trim whitespace from either end of line
	set line [string trim $line]

	# Parse lines of file
	switch -regexp $line {

	    {^(\#.*|::|^$)} {
		# Comment line or specifier line - ignore
	    } 

	    {^[^ \|]+.*:.} {
		# New record: Characters followed by a ':' and more characters
		# If there's an existing record accumulated, push it out
		if {[string length $record] > 0} {
		    set index [incr index]
		    processRecord $dbNS $record scalar $keyConcatenator $index
		    set record ""
		}
		# Start accumulating a new record
		set record $line
	    } 

	    {^\|[^\|]} {
		# Continuation line:  a '|' and then text
  		if {[string length $record] == 0} {
		    set errMess "ERROR(readDb): Unexpected record\
                        continuation line:\n    $line"
		    return -code error $errMess
		}
		set record ${record}[string range $line 1 end] 
	    }
	}
    }
    close $dbid
    # If there's a last record accumulated, push it out
    if {[string length $record] > 0} {
        set index [incr index]
	processRecord $dbNS $record scalar $keyConcatenator $index
	set record ""
    }    

    # Output of database information
    if {$debug && [set ${dbNS}::debug]} {
	dumpDbVal $dbNS
    }

    return 1
}


# CODEINFO BEGIN ######################################################

# info proc configureDb> Given the attribute line for a database,
#     initialize the open database namespace.

# arg dbNS> Name of the namespace containing the database.
# arg dbClassVal> The class ofthe  database (main, param, const).
# arg dbFieldList> List of fields in the database (with key and auto 
#     annotations).
# arg auxDbListVal> Value to set ${dbNS}
# return> Key.

proc webSubmit::database::configureDb {dbNS dbClassVal dbFieldList {auxDbListVal {}}} {
    variable debug
    alias $dbNS attrList auxDbList dbClass dbKeyList fieldList keyList\
        numAutoField numField 

    catch {unset keyList attrList}

    if {$debug} {
	puts "\nTRACE(configureDb): Configuring database parameters in \
            namespace $dbNS"
    }

    set dbClass $dbClassVal
    set fieldList $dbFieldList
    set dbKeyList {}

    # Compute numField and numAutoField
    set numField [llength $fieldList]
    set numAutoField 0
    foreach item $fieldList {
        if [regexp "^(key:)?auto\$" $item] {
	    set numAutoField [incr numAutoField]
	}
    }  

    # Split field descriptor into key and attribute lists
    foreach item $fieldList {
	if [regexp {^key:([^ ]+)$} $item match keyName] {
	    lappend keyList $keyName
	} else {
            lappend attrList $item
	}
    }

    # Set auxDbList
    set auxDbList $auxDbListVal

    if {$debug || [set ${dbNS}::debug]} {
        dumpDbState $dbNS
    }  
    return $dbNS
}

# CODEINFO BEGIN #######################################################

# info proc writeDb>  Process a database namespace $dbNS to produce a 
#     database *.db file.

# arg dbNS>  Name of the database namespace.
# arg keyConcatenator> Character used to separate different fields in a
#     key.
# arg pathName> Path to database file
# return> Status.
# loc-sclr line> The buffer in which to accumulate the record as a string.

proc webSubmit::database::writeDb {dbNS keyConcatenator {pathName ""}} {
    variable debug
    alias $dbNS attrList dbClass dbFile dbKeyList fieldList keyList auxDbList

    if {"$pathName" == ""} {
	set outDbFile $dbFile.db
    } else {
	set outDbFile $pathName
    }

    set fileId [open $outDbFile w]
    tclChmod 664 $outDbFile

    # Write out ::RECORD_ATTRIBUTES:: line
    puts $fileId "::RECORD_ATTRIBUTES:: [join $fieldList]"

    # Write out ::AUXILIARY_DATABASES:: line
    if {[info exists auxDbList] && [llength $auxDbList] != 0} {
	puts $fileId "::AUXILIARY_DATABASES:: [join $auxDbList]"
    }

    # Write out ::DATABASE_CLASS:: 
    puts $fileId "::DATABASE_CLASS:: $dbClass"

    # For all keys in the database write out their values
    foreach theKey $dbKeyList {
	set line ""
	set keyValList [trimSplit $theKey $keyConcatenator]
	foreach theFldName $fieldList {
	    if [string match *auto $theFldName] {
		# Skip auto fields
		continue
	    } elseif [string match key* $theFldName] {
		# key fields
		set fld [string trim $theFldName]
		regsub "key:" $fld "" fld
		set listIndex [lsearch $keyList $fld]
		set addition [lindex $keyValList $listIndex]
	    } else { 
		# attribute fields
		set addition [getDbVal $dbNS $theKey $theFldName]
		if {"$addition" == ""} {
		    set addition "*"
		}
	    }
	    if [regexp ":" $addition] {
		set addition "\"$addition\""
	    }
	    if {"$line" == ""} {
		set line $addition
	    } else {
		set line "$line : $addition"
	    }
	}
	puts $fileId $line
    }

    close $fileId
}


# CODEINFO BEGIN #######################################################

# info processRecord>  Process database record input format to produce
#     entries in DB.  

# arg dbNS>  Name of the database namespace.
# arg record>  String representing a colon-delimited database
#     input record.
# arg storageMode>  Either scalar or list.  
#     Scalar means set the element of the database to the value; 
#     list means append the value to what is there.
# arg keyConcatenator>  Character to separate key fields in the combined key.
# arg index>  Number of record (0, 1, ...).  
# return> key of record in database

# loc-list allfldsList>  List of all fields for a given entry.
# loc-list fieldList>  List of all field names in each record of the 
#     database.
# loc-sclr attr> Acumulated attribute string.
# loc-sclr key> Accumulated key string.
# loc-sclr numAutoField>  Number of field names that are auto and so not
#     represented in actual records.
# loc-sclr numField>  Number of field names in fieldList.
 
proc webSubmit::database::processRecord {dbNS record storageMode\
    {keyConcatenator "+"} {index missing}} {
    variable debug
    alias $dbNS attrList dbClass dbKeyList fieldList numAutoField numField
    variable internalConcatenator

    # Convert record to list, check length
    set allFldList [trimSplit $record ":"]
    set allFldListLength [llength $allFldList]
    set netNumField [expr $numField - $numAutoField]
    if {$allFldListLength != $netNumField} {
	set errMess "ERROR(processRecord): Expected number of fields\
            for record is $netNumField;\
            \n    number encountered is $allFldListLength\
            \n    the record is \"$record\""
	return -code error $errMess
    }

    # Extract and concatenate the key and attribute fields, inserting any auto
    # fields that are required
    set key ""
    set attr ""
    set fldIndex -1
    foreach item $fieldList {
        # Supply value of next field, checking for auto field
	if [string match *auto $item] {
            # Auto-indexing field; supply record index.
	    set fld " [format "%05d" $index] "
	} else {
	    # Explicit field, supply next field from input
	    set fldIndex [incr fldIndex]
	    set fld [string trim [lindex $allFldList $fldIndex]]
            if {"$fld" == "*"} {
		set fld ""
	    }
	}

	# Check for key field
	if [string match key* $item] {
	    # First, strip white space from around key and replace internal
	    # white space with the internalConcatenator character.
	    set fld [string trim $fld]
	    regsub -all "\[ \t\]+" $fld $internalConcatenator fld
	    # Key field - concatenate to key
	    if {"$key" == ""} {
		set key $fld
	    } else {
		set key [joinKey [list $key $fld] $keyConcatenator]
	    }
	} else {
	    if {$attr == ""} {
		set attr $fld
	    } else {
		set attr [join [list $attr $fld] :]
	    }
	    lappend attrValList [string trim $fld]
	}
    }

    # Store the attribute list at entry key in the database
    
    switch $dbClass {

	main {
	    if [existDbVal $dbNS $key] {
		# Duplicate entries in database
		if {$debug} {
		    set errorMsg "WARNING(processRecord): Entry already\
                            exists in database ${dbNS} for key $key"
		    puts $errorMsg
		}
	    }

	    # Store entry in database
	    setDbVal $dbNS $attr $key 

	    # Create subsidiary key,fieldName => fieldValue entries  
	    foreach theAttrVal $attrValList theAttr $attrList {
		setDbVal $dbNS $theAttrVal $key $theAttr
	    }
	}

	param -

	const {
	    # Store entry in database
	    switch -exact $storageMode {
		scalar {
		    if {[existDbVal $dbNS $key]} {
			# Duplicate entries in database
			if {$debug} {
			    set errorMsg "WARNING(processRecord): Entry\
                                already exists in database ${dbNS} for key\
                                $key"
			    puts $errorMsg
			}
		    }
		    setDbVal $dbNS $attr $key 
		    setDbVal $dbNS $attr $key [lindex $attrList 0]
		}

		list {
		    if {![existDbVal $dbNS $key]} {
			set currentValList {}
		    } else {
			set currentValList [getDbVal $dbNS $key]
		    }
		    set newValue [lappend currentValList $attr]
		    setDbVal $dbNS $newValue $key
		    setDbVal $dbNS $newValue $key [lindex $attrList 0]
		}
	    }
	}
    }

    # Accumulate the list of keys for the database
    if {[lsearch -exact $dbKeyList $key] == -1} {
	lappend dbKeyList $key 
    }
    return $key
}


# CODEINFO BEGIN #######################################################

# info readTxt>   Text read procedure: Reads text regions from a file.  
#    Each region is delimited by a key value that corresponds to a key 
#    from another database

# arg dbNS> Namespace for text data
# return> Status (Error = 1)
# loc-sclr txtFile> File containing database text
# loc-arr DBTXT> Name of array containing text data (equivalent to
#     $txtvarName in $dbNS)
# loc-sclr> currentKey: Tag associated with current text
# loc-sclr> txtid: Filehandle for text file
# loc-sclr> line: Current line of data from file

proc webSubmit::database::readTxt {dbNS} {
    # Variable aliases
    alias $dbNS DBTXT DB dbKeyList txtFile debug

    # Text file may not exist for this database
    if {$txtFile == ""} {
	return
    }

    # Check for existence of text file
    if {![fileUsable $txtFile]} {
	set errorMsg  "ERROR(readTxt): File $txtFile does not exist\
	    or is not readable"
	return -code error $errorMsg
    }

    # Open file and process data
    set currentKey ""
    set txtid [open "$txtFile" r]
    if {$debug} {
	puts "\n\n"
	puts [header [list "Database Text for $dbNS:"] \# -]
    }
    while {[gets $txtid line] >= 0} {
	# Parse lines of file
	if [regexp {^(#.*)} $line match] {
	    # Comment line - Ignore
	} else {
	    if [regexp {^:: *KEY *:: *(.*)} $line match key] {
		set currentKey $key
		set newline 1
		if {$debug} {
		    puts [header [list "Key: $key"] - =]
		}
	    } elseif {$line != "" && $currentKey != ""} {
		if {$newline} {
		    append DBTXT($currentKey) "$line"
		} else {
		    append DBTXT($currentKey) " $line\n"
		}
		set newline 0
		if {$debug} {
		    puts "\"$line\""
		}
	    }
	}
    }
    close $txtid

    # Add text entries to main database if keys match
    foreach key [array names DBTXT] {
	if {[lsearch -exact $dbKeyList $key] != -1} {
	    set DB($key,_text) $DBTXT($key)
	}
    }

    return 0
}


# CODEINFO BEGIN #######################################################

# info getField>   Procedure to extract fields from list with separators
#     Leading and trailing white space is stripped from each element

# arg entry> String to extract
# arg ifs> Internal Field Separator for entry
# return> List corresponding to input string
# loc-sclr ifs> Internal Field Separator for entry
# loc-sclr entry> String to extract
# loc-list newlist_L> List of field entries

proc webSubmit::database::getField {entry ifs} {

    # Return entry unchanged if it does not contain ifs
    #  Use '\' in pattern to escape field separators with special
    #  meaning to regexp or Tcl interpreter
    set pattern "\\$ifs"
    if ![regexp "$pattern" $entry match] {
	return $entry
    }

    # Split and then strip leading and trailing white-space for each element
    foreach elt [trimSplit $entry $ifs] {
	lappend newList [string trim $elt]
    }

    return $newList
}


# CODEINFO BEGIN #######################################################

# info nameDbNS> Derive the namespace of a database from the pathname by
#     computing the tail of the root of the pathname.   

# arg dbPathName> Pathname to the database file.  
# return> Status

proc webSubmit::database::nameDbNS {dbPathName} {
    variable debug

    set dbNS [file tail [file rootname $dbPathName]]

    if {"$dbNS" == ""} {
	set errMess "ERROR(nameDbNS): Namespace name derived from\
            \"$dbPathName\" is null."
	return -code error \$errMess
    } else {
	return $dbNS
    }
}


# CODEINFO BEGIN #######################################################

# info openDb> Instantiate a database namespace.

# arg dbNS>  Name of the namespace to contain the database
# arg dbFile>  Name of the file defining the database
# arg reqAttrList>  List of required non-ley attributes (fields)
# arg-sclr NSDebug>  Debug value for the namespace
# return> Status

proc webSubmit::database::openDb {dbNS {dbFile ""} {reqAttrList {}} {NSDebug 0}} {
    variable debug
    
    if {$debug} {
	puts "\nTRACE(openDb): Opening namespace $dbNS."
    }

    # Make namespace path for dbNS - relative and absolute cases
    if {![string match "::*" "$dbNS"]} {
        set callingNS [uplevel 1 namespace current]
        set openNS ${callingNS}::${dbNS}
        regsub "::::" $openNS "::" openNS
    } else {
        set openNS $dbNS
    }

    namespace eval $openNS {
        set locDebug [uplevel "set NSDebug"]
	variable debug $locDebug

	# Namespace definitions
	variable thisNS [namespace current]
	variable upNS [namespace parent]

        if {$debug} {
            puts "TRACE($thisNS): Performing namespace eval."
	}
        
        # Access to variables in the webSubmit root namespace
        alias ::webSubmit wsLibDir

	# Database variables
	variable DB
	variable DBTXT
  	variable dbFile 
	variable txtFile 
	variable fieldList
	variable numField
	variable numAutoField
        variable keyList
	variable attrList
	variable reqAttrList 
	variable dbKeyList

	set locFile [uplevel "set dbFile"]
        set fileRoot [file rootname $locFile]
	set dbFile $fileRoot
        if {[fileUsable $fileRoot.txt]} {
	    set txtFile $fileRoot.txt
	}
	
	set locReqAttrList [list [uplevel "set reqAttrList"]]
	if {[llength $locReqAttrList] != 0} {
	    set reqAttrList $locReqAttrList
	}	
    }

    if {$debug || [set ${openNS}::debug]} {
        dumpDbState $dbNS
    }  
    return $dbNS
}


# CODEINFO BEGIN #######################################################

# info serializeDb> Serialize a database namespace.
# arg dbNS> Name of the namespace to contain the database
# arg pathName> Path to serialized file
# return> None

proc webSubmit::database::serializeDb {dbNS {pathName ""}} {
    variable debug
    alias $dbNS DB DBTXT dbClass dbFile txtFile fieldList numField\
        numAutoField keyList attrList reqAttrList dbKeyList auxDbList

    set varList {DB DBTXT dbClass dbFile txtFile fieldList numField\
        numAutoField keyList attrList reqAttrList dbKeyList auxDbList}

    if {"$pathName" == ""} {
	set tclFile $dbFile.tcl
    } else {
        set extension [file extension $pathName]
	if {$extension != ""} {
	    set errMess "ERROR(serializeDb): Should not specify extension\
                 $extension with $pathName."
	    return -code error $errMess
        } else { 
            set tclFile $pathName
	}
    }
    tclTouch $tclFile 
    if {$debug} {
        puts "\nTRACE(serializeDb): The database namespace is $dbNS.\
            \n    Writing to file $tclFile."
    }
    serialize $dbNS $varList $tclFile

    return $dbNS
}


# CODEINFO BEGIN #######################################################

# info proc setDbVal> Given a database namespace, a key, and a field, 
#     set the value of that field.

# arg dbNS> Name of the namespace containing the database.
# arg key> Key of the record of interest.
# arg fieldName> Name of the field of interest.
# arg value> Value of entry in database
# return> Error status (0 = no error)

proc webSubmit::database::setDbVal {dbNS value key {fieldName ""}} {

    if [nsExist $dbNS] {
	if {[string compare $fieldName ""] == 0} {
	    set subkey $key
	} else {
	    set subkey "$key,$fieldName"
	}

	set ${dbNS}::DB($subkey) $value
	return 0
    } else {
	set errMsg "ERROR(setDbVal): Database namespace $dbNS does not exist."
	return -code error $errMsg
    }
}

# CODEINFO BEGIN ######################################################

# info proc getDbVal> Given a database namespace, a key, and a field, 
#     return the value of that field.

# arg dbNS> Name of the namespace containing the database.
# arg key> Key of the record of interest.
# arg fieldName> Name of the field of interest.
# return> Value of field

proc webSubmit::database::getDbVal {dbNS key {fieldName ""}} {
    variable debug

    if [nsExist $dbNS] {
	if {[string compare $fieldName ""] == 0} {
	    set subkey $key
	} else {
	    set subkey "$key,$fieldName"
	}

        if [info exist ${dbNS}::DB($subkey)] {
	    set retrieval [set ${dbNS}::DB($subkey)]
	    if {[llength $retrieval] == 1} {
		set retrieval [join $retrieval]
	    }
	    return $retrieval
        } else {
	    set errMsg "ERROR(getDbVal): Field $subkey is not\
                defined in ${dbNS}::DB."
	    return -code error $errMsg
	}
    } else {
	set errMsg "ERROR(getDbVal): Database namespace $dbNS does not\
            exist."
	return -code error $errMsg
    }
}


# CODEINFO BEGIN ######################################################

# info proc existDbVal> Given a database namespace, a key, and a field, 
#     determine if the field exists.

# arg dbNS> Name of the namespace containing the database.
# arg key> Key of the record of interest.
# arg fieldName> Name of the field of interest.
# return> True (1) or false (0).

proc webSubmit::database::existDbVal {dbNS key {fieldName ""}} {
    variable debug

    if [nsExist $dbNS] {
	if {[string compare $fieldName ""] == 0} {
	    set subkey $key
	} else {
	    set subkey "$key,$fieldName"
	}

        if [info exist ${dbNS}::DB($subkey)] {
	    return 1
        } else {
	    return 0
	}
    } else {
	set errMsg "ERROR(existDbVal): Database namespace $dbNS does not\
            exist."
	return -code error $errMsg
    }
}


# CODEINFO BEGIN #######################################################

# info dumpDbVal>  Print current state of DB array.  
# arg dbNS>  Name of the database namespace.
# return> None.
 
proc webSubmit::database::dumpDbVal {dbNS} {
    alias $dbNS DB dbKeyList

    puts "\n"
    puts [header [list "Record Key to Value Mapping for Namespace $dbNS"] \# -]
    if {[llength $dbKeyList] == 0} {
	puts "    No entries in the record key to value mapping."
    } else {
	foreach item $dbKeyList {
	    puts "    DB($item) = $DB($item)"
	    foreach subitem [array names ${dbNS}::DB $item,*] {
		puts "        DB($subitem) = $DB($subitem)"
	    }
	    puts "\n"
	}
    }
    return 0
}


# CODEINFO BEGIN #######################################################

# info dumpDbState>  Print current state of database.  
# arg dbNS>  Name of the database namespace.
# return> None.
 
proc webSubmit::database::dumpDbState {dbNS} {
    alias $dbNS attrList dbFile fieldList keyList numAutoField numField \
	reqAttrList auxDbList dbKeyList

    puts "\n"
    puts [header [list "Database State Information for Namespace $dbNS:"] \
	      \# -]
    if [info exists dbClass] {
	puts "  Database class:  $dbClass"
    }
    if [info exists numField] {
	puts "  Number of fields:  $numField"
    }
    if [info exists numAutoField] {
	puts "  Number of auto fields:  $numAutoField"
    }
    if [info exists fieldList] {
	puts "  Fields:  $fieldList"
    }
    if [info exists keyList] {
	puts "  Number of key fields:  [llength $keyList]"
    }
    if [info exists keyList] {
	puts "  Key fields:  $keyList"
    }
    if [info exists attrList] {
	puts "  Number of attributes: [llength $attrList]"
    }
    if [info exists attrList] {
	puts "  Attribute fields: $attrList"
    }
    if [info exists reqAttrList] {
	puts "  Required attribute fields:  $reqAttrList"
    }
    if [info exists dbFile] {
	puts "  Database file root:  $dbFile"
    }
    if [info exists auxDbList] {
	puts "  Auxiliary database list:  $auxDbList"
    }    
    if [info exists dbKeyList] {
	puts "  Database key root:  $dbKeyList"
    }    
    return $dbNS
}


# CODEINFO BEGIN ######################################################

# info proc splitKey> Convert $key to list of key attribute values.
# arg key> The key.
# arg keyConcatenator> The character to separate key fields in a key.
# return> The list of key attribute values.

proc webSubmit::database::splitKey {key\
    {keyConcatenator ""}} {
    if {"$keyConcatenator" == ""} {
	set keyConcatenator $webSubmit::database::stdConcatenator
    }
    return [split $key $keyConcatenator keyConcatenator]
}

# CODEINFO BEGIN ######################################################

# info proc joinKey> Make a key from a list of key fields.
# arg keyList> A list of key field values to be made into a key.
# arg keyConcatenator> The character to separate key fields in a key.
# return> The concatenated key.  

proc webSubmit::database::joinKey {keyList\
    {keyConcatenator ""}} {
    if {"$keyConcatenator" == ""} {
	set keyConcatenator $webSubmit::database::stdConcatenator
    }
    return [join $keyList $keyConcatenator]
}























