# WebSubmit 2.0 wsForms.tcl @(#)wsForms.tcl	1.21 SCCS/s.wsForms.tcl  99/07/12 

# AUTHORS #############################################################

# WebSubmit 2.0:
#  Authors: John Koontz, Ryan McCormack, Judith Devaney
#  Original Development Work: Robert Lipman, John Koontz, Katherine
#      Pagoaga

# Copyright Information ###############################################

#   This software was developed at the National Institute of Standards
#   and Technology by employees of the Federal Government in the course
#   of their official duties.  Pursuant to title 17 Section 105 of the
#   United States Code this software is not subject to copyright
#   protection and is in the public domain.

# CODEINFO BEGIN ######################################################

# info package wsForms> Produce a WebSubmit form page.

# wsForms-sclr currentHostName> Name of the current host.
# wsForms-sclr currentModuleName> Name of the current module.
# wsForms-sclr currentModeName> Name of the current mode.
# wsForms-sclr instanceName> Name of the current element instance.
# wsForms-sclr elementName> Name of the current element.
# wsForms-sclr numFormCols> Number of columns the current form is wide.
# wsForms-sclr numCols> Number of columns so far in the current form 
#     row.
# wsForms-sclr debug> If true, run debug code.

# ???

# CODEINFO END ########################################################

# Package Declarations ################################################
package provide wsForms 1.0

# Namespace Definition ################################################
namespace eval webSubmit::forms {
    # Namespace definitions
    variable thisNS [namespace current]
    variable upNS [namespace parent]
    variable wsNS [namespace parent]

    # Namespace imports
    catch {namespace import ${wsNS}::util::*}
    catch {namespace import ${wsNS}::database::*}
    catch {namespace import ${wsNS}::formElements::*}
    catch {namespace import ${wsNS}::auth::*}

    # Namespace exports
    namespace export formPageBuilder
    namespace export checkUserModule
    namespace export getFormDir    

    # Variable definitions
    variable numFormCols 
    variable numRowCols
    variable currentHostName
    variable currentModuleName 
    variable currentModeName 
    variable marginBgColor "burlywood"
    variable marginFgColor "black"
    variable bodyBgColor "cornsilk"
    variable bodyFgColor "black"
    variable titleBgColor "cornflowerblue"
    variable titleFgColor "black"
    variable tableCellSpacing 0
    variable tableBorder 0
    variable instanceName 
    variable elementName
    variable debug 0
    variable userSessName "session"
    alias $wsNS wsUserDir wsLibDir wsModuleDir userPath moduleHead

    # Namespace for session information
    namespace eval session {
	# Namespaces
	variable upNS [namespace parent]
	variable thisNS [namespace current]
	variable wsNS [set [namespace parent]::wsNS]

	# Namespace imports
	catch {namespace import ${wsNS}::util::*}
	catch {namespace import ${wsNS}::database::*}
	catch {namespace import ${wsNS}::formelements::*}
	catch {namespace import ${wsNS}::auth::*}
	alias $upNS wsNS userPath wsLibDir

	# Namespace exports
	namespace export getSessionData
	namespace export getCgiData
	namespace export session

	# General variable list
	variable sessCgiData ""
	variable sessType "default"
	variable userDefName "default"
	variable userLastName "lastSession"
	variable systemDefName "system"
	variable loadedSessName "None"

	# Database variables
	variable DB
	variable dbKeyList
	variable attrList
	variable reqAttrList

        variable debug 1
    }

    if {$debug} {
	puts "TRACE(wsForms.tcl):  Creating ::webSubmit::forms"
    }
}

# Procedure Definitions ###############################################

# CODEINFO BEGIN ######################################################
# info proc session> Procedure to perform session-related tasks, such
#     as loading and saving.
#
# Syntax:
#   session (load|save|delete) [--default] [-vars sessVarList]
#           [-mode modeName] [-name sessionName] moduleName hostName
#
#     load: Load a saved session
#     save: Save the current module input as a session
#     delete: Remove a session from the list of possible sessions
#
#   Additional arguments:
#     -vars sessVarList: List of variables to save in session file
#      (req. if command=save and form is not auto-generated)
#     -mode modeName: Specify modename (optional)
#     -name sessionName: Name of session to save or load 
#      (req. if command=save or delete)
#     --default: Specifies default session for loading
#     moduleName: Name of the module for the given session (req.)
#     hostName: Name of the host to run the module on (req.)

# arg command> One of load, save, delete
# arg args> Additional arguments
# return> Depends on command

proc webSubmit::forms::session::session {command args} {
    variable upNS
    alias $upNS debug

    # Process "command" argument
    set commandList [list load save delete]
    if {[lsearch -exact $commandList $command] == -1} {
	return -code error "ERROR(session): Invoked with invalid command"
    }

    # Process additional arguments
    set ARGV(modeName) "BASIC"
    set ARGV(name) ""
    set ARGV(vars) ""
    set ARGV(default) 0
    eval argParse $args
    set moduleName [lindex $ARGV(argList) 0]
    set hostName [lindex $ARGV(argList) 1]
    set sessName $ARGV(name)
    set sessVarList $ARGV(vars)
    set default $ARGV(default)

    if {$moduleName == ""} {
	# Invalid invocation
	return -code error "ERROR(session): Invoked without moduleName"
    }

    if {$hostName == ""} {
	# Invalid invocation
	return -code error "ERROR(session): Invoked without hostName"
    }    

    if {$debug} {
	puts "  Session Task Manager: $command"
	puts "  [array get ARGV]"
    }

    # Process commands
    if {$command == "load"} {
	if {$sessName != ""} {
	    loadSession $default $moduleName $hostName $sessName
	} else {
	    loadSession $default $moduleName $hostName
	}
    } elseif {$command == "save"} {
	if {$sessVarList != "" && $sessName != ""} {
	    # Copy Session data into session database
	    vars2db $sessVarList

	    # Save session data
	    saveSession $moduleName $hostName $sessName
	}
    } elseif {$command == "delete"} {
	deleteSession $moduleName $hostName $sessName
    }
    return 0
}


# CODEINFO BEGIN ######################################################

# info proc getSessFileName> Procedure to choose the session file to
#     load when a form is invoked, if any.  If no appropriate session
#     file can be located from the user or system trees, then the
#     procedure returns a null filename.
#
# Precedence rules for determining session file to load:
#    1. User session settings
#        --> Determine preference for "default" or "last"
#           (i) Default: If unavailable, load system
#           (ii) Last: If unavailable, load user default
#    2. System defaults
#    3. No session information

# arg moduleName> Name of the module for which state information is needed
# arg hostName> Name of the host for which state information is needed
# return> Session filename (full path) or "" if no file can be located

proc webSubmit::forms::session::getSessFileName {moduleName hostName} {
    variable sessType
    variable upNS
    variable userDefName
    variable userLastName
    variable systemDefName
    variable loadedSessName
    alias $upNS debug

    # Various session files
    set userModDir [${upNS}::getFormDir -module $moduleName $hostName]
    set userSessDir [${upNS}::getFormDir -session $moduleName $hostName]
    set systemDir [${upNS}::getFormDir -system $moduleName $hostName]
    set userDef [file join $userSessDir "$userDefName.tcl"]
    set userLast [file join $userSessDir "$userLastName.tcl"]
    set systemDef [file join $systemDir "$systemDefName.tcl"]

    # Existence checks for various files and directories
    foreach file [list $userModDir $userSessDir $userDef $userLast $systemDir\
        $systemDef] {
	set EXIST($file) [file exists $file]
    }

    # Determine whether user has session data
    if {$EXIST($userSessDir)} {
	if {$EXIST($userDef)} {
	    set loadedSessName $userDefName
	    set fileName $userDef
	} else {
	    if {$EXIST($userLast)} {
		set loadedSessName $userLastName
		set fileName $userLast
	    } else {
		if {$EXIST($systemDef)} {
		    set loadedSessName $systemDefName
		    set fileName $systemDef
		} else {
		    set fileName ""
		}
	    }
	}
    } elseif {$EXIST($systemDir)} {
	# Search for system defaults
	if {$EXIST($systemDef)} {
	    set loadedSessName $systemDefName
	    set fileName $systemDef
	} else {
	    set fileName ""
	}
    } else {
	set fileName ""
    }

    if {$debug} {
	puts "Session File Name ($moduleName, %hostName): $fileName"
    }
    return $fileName
}



# CODEINFO BEGIN ######################################################

# info proc checkUserModule> Procedure to create a user directory for
#     a specified module, and to set up default values for that
#     module.

# arg moduleName> Name of module to initialize
# arg hostName> Name of the host to initialize
# return> None

proc webSubmit::forms::checkUserModule {moduleName hostName} {
    set userSessDir [getFormDir -session $moduleName $hostName]
    if ![file exists $userSessDir] {
	# Create user module and session directories
	file mkdir $userSessDir
    } else {
	if ![file isdirectory $userSessDir] {
	    return -code error "ERROR(checkUserModule): Cannot create\
                  user module directory for module \"$moduleName\"\
                  and host \"hostName\" - File exists"
	}
    }
    return
}


# CODEINFO BEGIN ######################################################

# info proc loadSession> Procedure to load session data into an array
#     in the forms namespace called SESSION

# arg default> Determines whether to load system defaults
# arg moduleName> Name of module
# arg hostName> Name of host
# arg sessName> Name of session to load
# return> None

proc webSubmit::forms::session::loadSession {default moduleName hostName\
    {sessName ""}} {
    variable DB
    variable upNS
    variable systemDefName
    variable loadedSessName
    alias $upNS debug
    catch {unset DB}

    if {$debug} {
	cgi_http_head
	puts "  Loading session \"$sessName\" for \"$moduleName\" on \"$hostName\""
    }

    # Set paths and session file name
    if {[isAdmin] || $default} {
	set dirName [${upNS}::getFormDir -system $moduleName $hostName]
    } else {
	set dirName [${upNS}::getFormDir -session $moduleName $hostName]
    }

    if {$sessName == ""} {
	if {$default} {
	    # Load system defaults
	    set sessFileName [file join $dirName "$systemDefName.tcl"]
	    set tmpName $systemDefName
	} else {
	    set sessFileName [getSessFileName $moduleName $hostName]
	    set tmpName [file rootname [file tail $sessFileName]]
	}
    } else {
	set sessFileName [file join $dirName "$sessName.tcl"]
	set tmpName $sessName
    }

    if {!($sessFileName == "") && [file exists $sessFileName]} {
	set loadedSessName $tmpName
	# Load serialized session file
	if [loadSerial $sessFileName] {
	    catch {unset DB}
	    set loadedSessName "None"
	} else {
	    if {$debug} {
		puts "Session Loaded: \"$loadedSessName\" ($sessFileName)"
	    }
	}
    }
    return
}


# CODEINFO BEGIN ######################################################

# info proc saveSession> Procedure to save a named session for a given
#    module.  It is assumed that all session data has already been
#    placed in the forms array SESSION

# arg moduleName> Name of module
# arg hostName> Name of host
# arg sessName> Name of session to save
# return> None

proc webSubmit::forms::session::saveSession {moduleName hostName sessName} {
    variable thisNS
    variable upNS
    variable loadedSessName
    alias ${upNS} debug
    
    if {$debug} {
	puts "   Saving session \"$sessName\""
    }

    # Set paths and session file name
    if [isAdmin] {
	set dirName [getFormDir -system $moduleName $hostName]
    } else {
	set dirName [getFormDir -session $moduleName $hostName]
    }
    set sessFileName [file join $dirName "$sessName.tcl"]

    # Verify session directory
    if ![file isdirectory $dirName] {
	cgiBomb saveSession "Invalid directory \"$dirName\" for session"
    }

    if {$debug} {
	puts "   Serializing \"$sessName\" into $sessFileName"
    }

    # Serialize session data
    set loadedSessName "$sessName"
    set varList [list DB loadedSessName]
    serialize $thisNS $varList $sessFileName
}
 
   
# CODEINFO BEGIN #####################################################

# info proc deleteSession> Procedure to delete a named session
# arg sessFileName> Name of session file to delete
# return> None

# arg moduleName> Name of module
# arg hostName> Name of host
# arg sessName> Name of session to be deleted
# return> None
proc webSubmit::forms::session::deleteSession {moduleName hostName sessName} {
    # Set paths and session file name
    if [isAdmin] {
	set dirName [getFormDir -system $moduleName $hostName]
    } else {
	set dirName [getFormDir -session $moduleName $hostName]
    }
    set sessFileName [file join $dirName "$sessName.tcl"]

    if [catch {file delete $sessFileName} err] {
	cgiBomb deleteSession "File could not be deleted ($err)"
    }
    return
}

# CODEINFO BEGIN ######################################################

# info proc makeDefSession> Procedure to build system default values for a 
#     session from a form metadatabase or from existing system defaults.
#     Only valid admin users can build system-wide default sessions.
#     CURRENT STATUS: Unused.

# arg moduleName> Name of module that needs default session file
# arg hostName> Name of host that needs default session file
# return> None

proc webSubmit::forms::makeDefSession {moduleName hostName} {
    variable thisNS
    variable SESSION

    # Set directory and file locations
    set userDir [getFormDir -module $moduleName $hostName]
    set systemDir [getFormDir -system $moduleName $hostName]
    set sysSessFileName [file join $systemDir "session.tcl"]
    set userSessFileName [file join $userDir "session.tcl"]
    set mdbFileName [file join $systemDir "form.mdb"]
    set mdbSourceName [file join $systemDir "form.tcl"]

    if [isAdmin] {
	# Build system-wide defaults
	if [file isfile $mdbFileName] {
	    # Load metadatabase data
	    if [loadSerial $mdbSourceName] {
		# Error loading serial version of metadatabase
		#  Build the database from the metadatabase file
		# buildDb $moduleName
	    }
	} else {
	    # No metadatabase file exists for this form
	    return -code error "ERROR(updateDefSession): No form\
                  metadatabase exists for module \"$moduleName\"\
                  and host \"$hostName\""
	}
	
	# Tranfer appropriate data from form database into SESSION
	set formNameSpace [nsJoin $thisNS $moduleName]
	foreach key [set ${formNameSpace}::dbKeyList] {
	    set value [getDbVal $formNameSpace $key defaultValue]
	    set SESSION($key) $value
	}
	
	# Serialize system-wide session data for this module
	serialize $thisNS SESSION $sysSessFileName
    } else {
	# Copy system defaults to regular user's directory
	if [file exists $sysSessFileName] {
	    if [file isdirectory $userDirName] {
		file copy -force $sysSessFileName $userSessFileName
	    }
	}
    }
    return
}

# CODEINFO BEGIN ######################################################

# info proc getSessionNames> Procedure to return the names of all 
#     session names for a given module-host pair

# arg moduleName> Name of module
# arg hostName> Name of host
# return> List of session file names

proc webSubmit::forms::getSessionNames {moduleName hostName} {
    if ![isAdmin] {
	# Check for existence of user module directories
	checkUserModule $moduleName $hostName

	# Get session names from directory
	set userSessDir [getFormDir -session $moduleName $hostName]
	set sessNameList ""
	if [file isdirectory $userSessDir] {
	    if ![catch {glob "$userSessDir/*"} fileList] {
		foreach file [glob "$userSessDir/*"] {
		    lappend sessNameList [file tail [file rootname $file]]
		}
	    }
	}
    } else {
	set sessNameList "session"
    }
    return $sessNameList
}
	


# CODEINFO BEGIN ######################################################

# info proc getFormDir> Procedure to return the appropriate directory
#     name for the given module and user.  Present implementation:
#        Type:
#          -module: $userPath/$hostName/$moduleName
#          -session: $userPath/$hostName/$moduleName/$userSessionDir
#          -system: $wsModuleDir/$hostName/$moduleName

# arg type> -session = Return user session dir, 
#           -module = Return user module dir,
#           -system = Return system module dir
# arg moduleName> Name of module
# arg hostName> Name of host
# return> Directory name for either session files or module files

proc webSubmit::forms::getFormDir {type moduleName hostName} {
    variable wsNS
    variable userSessName
    variable userPath
    variable debug
    alias $wsNS wsModuleDir

    if {$hostName != "Unknown"} {
	switch -exact -- $type {
	    -system {set path [file join $wsModuleDir $hostName $moduleName]}
	    -module {set path [file join $userPath $hostName $moduleName]}
	    -session {set path [file join $userPath $hostName $moduleName $userSessName]}
	    default {set path [file join $userPath $hostName $moduleName]}
	}
    } else {
	return -code error "The host \"$hostName\" is unknown."
    }

    if {$debug} {
	puts "   Form Directory lookup ($moduleName,$type): $path"
    }
    return $path
}


# CODEINFO BEGIN ######################################################

# info proc moduleIsAuto> Procedure to determine whether a module is
#     auto-generated from a database, or based on a hand-written
#     CGI script

# arg moduleName> Name of the module to consider
# arg hostName> Name of the host to consider
# return> 1 = Auto, 0 = Manual
    
proc webSubmit::forms::moduleIsAuto {moduleName hostDir} {
    set moduleDir [getFormDir -system $moduleName $hostName]
    set moduleFile [file join $moduleDir "form.mdb"]
    if [file exists "$moduleFile"] {
	# Form is auto-generated from $moduleFile
	return 1
    } else {
	return 0
    }
}


# CODEINFO BEGIN ######################################################

# info proc vars2db> Store a list of CGI input variables in the session 
#     database.  A protected version of the variables is extracted from the
#     global array element _cgi(uservars) and stored in the session database.
#     This routine assumes that a cgi_input has already been performed to 
#     load _cgi(uservars) with the CGI input data.

# arg varList> List of variables to save
# return> None

proc webSubmit::forms::session::vars2db {varList} {
    variable DB
    global _cgi
    global _cgi_uservar

    foreach var $varList {
	if {[lsearch -exact $_cgi(uservars) $var] != -1} {
	    set DB($var) [string trim "$_cgi_uservar($var)"]
	}
    }
    return
}

# CODEINFO BEGIN ######################################################

# info proc getCgiData> Convert session database into a format that can be
#     submitted with the form as a query string

# arg> None
# return> String to use as CGI input

proc webSubmit::forms::session::getCgiData {} {
    variable DB
    variable sessCgiData

    foreach var [array names DB] {
	set value [getDbVal [namespace current] $var]
	set sessCgiData "$sessCgiData&[cgi_cgi_set $var $value]"
    }

    eval return \"$sessCgiData\"
}

# CODEINFO BEGIN ######################################################

# info proc getSessionData> Procedure to extract data for a given
#     session from the session database.  Data is converted from a
#     CGI-safe form back into the original form.

# arg> None
# return> None

proc webSubmit::forms::session::getSessionData {} {
    variable DB
    
    foreach var [array names DB] {
	uplevel "set $var \{[cgi_unquote_input $DB($var)]\}"
    }
    return
}
















































































