# WebSubmit 2.0 wsMaster.tcl @(#)wsMaster.tcl	1.26 SCCS/s.wsMaster.tcl  99/07/12 

# AUTHORS #############################################################

#  Authors: John Koontz, Ryan McCormack, Judith Devaney
#  Original Development Work: Robert Lipman, John Koontz, Katherine
#      Pagoaga

# Copyright Information ###############################################
 
# This software was developed at the National Institute of Standards
# and Technology by employees of the Federal Government in the course
# of their official duties.  Pursuant to title 17 Section 105 of the
# United States Code this software is not subject to copyright
# protection and is in the public domain.

# CODEINFO BEGIN ######################################################

# info package wsMaster> This package sets up the ::webSubmit::master
#     namespace and defines a series of procs used to process the
#     master.db and its cgi-generated Web page representation, which
#     is the master page linking together all WebSubmit functions.

# webSubmit::master-list nodeScriptList > List of all webSubmit scripts
# webSubmit::master-list masterList > List of all master nodes in master page
# webSubmit::master-list childList > List of all children
# webSubmit::master-sclr debug> Debug master DB calls 
# webSubmit::master-sclr wsID> Websubmit user ID for current user
# webSubmit::master-arr DB> Entries in master page database
# webSubmit::master-arr DBTXT > Text in master page database
# webSubmit::master-sclr dbFile> File containing master database
# webSubmit::master-sclr txtFile> File containing master text
# webSubmit::master-sclr stateFile> File containing user state information
# webSubmit::master-sclr userPath> Path to user directories
# webSubmit::master-list reqAttrList > List of required master database 
#     attributes
# webSubmit::master-list attrList> List of actual master DB attributes
# webSubmit::master-list dbKeyList > List of all webSubmit nodes

# CODEINFO END ########################################################

# Package Declarations ################################################

package provide wsMaster 1.0

# Namespace Definition and Variable Initialization ####################

namespace eval webSubmit::master {
    # Procedures to be exported
    namespace export createMasterDb

    # Namespace definitions
    variable thisNS [namespace current]
    variable upNS [namespace parent]
    variable wsNS [namespace parent]

    # Procedures to be imported
    catch {namespace import ${wsNS}::database::*}
    catch {namespace import ${wsNS}::util::*}
    
    # Aliases variables
    alias $wsNS wsLibDir cgiHead userPath

    # General Variable definitions
    variable debug 0
    variable userConfigFile
    variable userUptoDate
    variable childList ""
    variable nodeScriptList
    variable masterList ""
    variable execError

    # Database variables
    variable DB
    variable DBTXT
    variable dbFile [file join $wsLibDir master]
    variable txtFile [file join $wsLibDir master.txt]
    variable reqAttrList [list visible state type title moduleName nodePath\
        defaultMode]
    variable attrList
    variable dbKeyList
    variable dbMade 0
}

# Procedure definitions ###############################################

# CODEINFO BEGIN ######################################################

# info proc createMasterDb> Procedure to create master database information.
#     Database file is parsed and all information is stored in a master 
#     namespace.

# return > Status

# webSubmit::master-list dbKeyList>  Database keys == Node identifiers
# webSubmit::master-list nodeScriptList>   Active node scripts
# webSubmit::master-list masterList> master node list
# loc-sclr thisNS> Name of master namespace

proc webSubmit::master::createMasterDb {} {
    variable wsLibDir
    variable dbMade
    variable userPath
    variable userUptoDate
    variable userConfigFile [file join $userPath "userConfig.tcl"]
    variable cgiHead
    set pathName [file join $wsLibDir master]

    # Variable declarations
    foreach var [list thisNS masterList dbKeyList DB debug] {
	variable $var
    }

    # Build generic master database
    buildDb $pathName $thisNS

    # Load user data if exists and is newer that the master database
    set masterFile [file join $wsLibDir "master.db"]
    if {[file exists $userConfigFile] && \
	    ([file mtime $userConfigFile] > [file mtime $masterFile])} {
	set userUptoDate 1
	loadSerial $userConfigFile
	if {$debug} {
	    puts "Loaded user configuration data from $userConfigFile"
	}
    } else {
	set userUptoDate 0
    }

    # Build link data for master web page
    variable nodeScriptList [defineLinks $thisNS $cgiHead]

    if {$debug} {
	puts "\n"
	puts [header [list "List of active nodes: $nodeScriptList"] \# =]
	foreach mod $nodeScriptList {
	    puts "  $mod = [cgi_link $mod]"
	}
    }

    # Analyze hierarchies for all entries
    foreach key $dbKeyList {
	parsePath [getDbVal $thisNS $key nodePath]
    }

    # Sort into index order
    set masterList [lsort -command compChild $masterList]
    foreach key $dbKeyList {
	if [existDbVal $thisNS $key children] {
	    set children [getDbVal $thisNS $key children]
	    set sortedChildren [lsort -command compChild $children]
	    setDbVal $thisNS $sortedChildren $key children 
	}
    }

    if {$debug} {
	puts "\n\n"
	puts [header [list "Parent-child relationships:"] \# =]
	foreach key $dbKeyList {
	    if [existDbVal $thisNS $key children] {
		puts "  Parent: $key   Children: [getDbVal $thisNS $key children]"
	    }
	}
    }

    # Update or build User configuration file if necessary
    if {!($userUptoDate) && $userPath != ""} {
	serialize $thisNS DB $userConfigFile
    }
    set dbMade 1
    return 0
}


# CODEINFO BEGIN ######################################################

# info proc parsePath>   Procedure to parse path and break it into its
#      components

# Example: path = parent/c1/gc1
#  - Adds 'parent' to list of masters (if not already present)
#  - Adds 'c1' to the children of 'parent'
#  - Adds 'gc1' to the children of 'c1'

# arg path> Node path to parse
# return> Status

# loc-arr DB> Master Database
# local-sclr parent> Parent node name
# local-sclr child> Child node name
# webSubmit::master-list masterList> List of 'master' nodes
# webSubmit::master-list childList> List of all 'child' nodes
# loc-sclr length> Number of elements in path
# loc-sclr end> Index of last entry in path

proc webSubmit::master::parsePath {path} {
    # Master namespace variables
    foreach var [list debug DB masterList childList thisNS] {
	variable $var
    }

    # Trim leading and trailing "/" from path for split to work
    set path [string trim $path "/"]

    # Split path into list of path elements
    set pathList [split $path "/"]
    
    # Add master to list if one of two conditions is met:
    #  1. masterList does not exist and node is "ON"
    #  2. masterList exists, node is "ON", and node is not already a
    #     member
    set master [lindex $pathList 0]
    if {![info exists masterList] || \
        ([info exists masterList] && ([lsearch -exact $masterList $master]\
        == -1))} {
	# Verify that master is a valid node
	if ![existDbVal $thisNS $master] {
	    set errorMsg  "ERROR (parse_path): \n  Element $master does not\
		    specify a valid database key"
	    return -code error $errorMsg
	} else {
	    # Verify that this node is visible and active
	    if {[getDbVal $thisNS $master visible] == "ON"\
                && [getDbVal $thisNS $master state] == "ON"} {
		# Add to list of master nodes
		lappend masterList $master
	    } else {
		# master element = OFF
		return 0
	    }
	}
    }
    
    # Determine pathlist paramaters
    set length [llength $pathList]
    set end [expr $length - 1]

    if {$debug} {
	puts "\nParsing path \"$path\" in NS $thisNS"
	puts "   Current PathList: $pathList"
	puts "   Master List: $masterList"
	puts "   Current Master: $master"
    }

    if {$length > 1} {
	# Element possesses children
	#  Loop over all children that can themselves be parents
	for {set i 0} {$i < $end} {incr i} {
	    # Find node_id key
	    set parent [lindex $pathList $i]
	    if ![existDbVal $thisNS $parent] {
		# Array data missing - Ignore this and all subsequent 
                # elements
                set errorMsg "WARNING (parse_path): \n  Element \"$parent\"\
                    in path \"$path\" is invalid"
		puts $errorMsg
		break
	    }
	    
	    # Determine immediate child of this id
	    set child [lindex $pathList [expr $i + 1]]
	    if ![existDbVal $thisNS $child] {
		# Array data missing - Ignore this and all subsequent 
                # elements
                set errorMsg "WARNING (parse_path): \n  Element \"$child\"\
                    in path \"$path\" is invalid"
		puts $errorMsg 
		break
	    }

	    # Construct lists only for visible and active children
	    if {[getDbVal $thisNS $child visible] == "ON"\
                && [getDbVal $thisNS $child state] == "ON"} {
		# Check to see if parent is visible and active.  
                # If either is OFF, then abort
		if {[getDbVal $thisNS $parent visible] == "ON"\
                    && [getDbVal $thisNS $parent state] == "ON"} {
		    # Check to see if child exists globally
		    if {![info exists childList] ||\
                        [lsearch -exact $childList "$child"] == -1} {
			# Check to see if child already exists for this parent
			if {$debug} {
			    puts "   Current parent: $parent"
			    puts "   Current child: $child"
			    puts "   Children? [existDbVal $thisNS $parent children]"
			}

			if {![existDbVal $thisNS $parent children]} {
			    # Create null child list
			    setDbVal $thisNS "" $parent children
			}

			set children [getDbVal $thisNS $parent children]
			if {[lsearch -exact $children "$child"] == -1} {
			    if {$debug} {
				puts "    Generating children..."
			    }
			    set children [lappend children "$child"]
			    setDbVal $thisNS $children $parent children
			    lappend childList "$child"
			}
		    } else {
			# Child has already been defined
			continue
		    }
		} else {
		    # Report error and abort
		    set errorMsg "WARNING (parse_path): \n  Element \"$child\"\
                    in path \"$path\" has an invisible or inactive parent."
		    puts $errorMsg 
		    break
		}
	    }
	}
    }
    return 0
}


# CODEINFO BEGIN ######################################################

# info proc makeCascade> Procedure to make a group of cascaded UL
#      entries in HTML

#  For a given list entry, the procedure generates a list
#   of all the children for this entry.  If any child has
#   children itself, then the procedure is invoked recursively.
#   In this way, a series of nested list entries can be generated
#
#  Example List:
#  -------------
#  <UL>
#     <LI> E1 (Parent of E1.1, E1.2, E1.3)
#         <UL>
#             <LI> E1.1 
#             <LI> E1.2 (Parent of E1.2.1, E1.2.2)
#                <UL>
#                    <LI> E1.2.1
#                    <LI> E1.2.2
#                </UL>
#             <LI> E1.3
#         </UL>
#
#     <LI> E2
#  </UL>

# arg parent> Parent node id from master database
# return> Status

# loc-arr DB> master database
# loc-list childList> Children of the input parent
# loc-sclr length> Number of children
# webSubmit::master-list nodeScriptList> List of active nodes (scripts)

proc webSubmit::master::makeCascade {parent} {
    # Master namespace variables
    foreach var [list thisNS DB nodeScriptList debug] {
	variable $var
    }

    if ![info exists DB] {
	set errorMsg "ERROR (makeCascade): \n  Invalid database name"
	return -code error $errorMsg
    }

    if {$debug} {
	puts "\n\n[header [list "Cascade Information"] # -]"
	puts "\nCascade for parent: $parent"
	if [existDbVal $thisNS $parent children] {
	    puts "   Children: [getDbVal $thisNS $parent children]"
	}
    }

    # Extract list of children
    if [existDbVal $thisNS $parent children] {
	set childList [getDbVal $thisNS $parent children]
    } else {
	return
    }
    set length [llength $childList]

    cgi_bullet_list {
	if {$length > 0} {
	    # Generate <LI> entries for all children
	    foreach child $childList {
		# Create HTML element
		if {[getDbVal $thisNS $child type] == "CONTAINER"} {
		    # Child is a container
		    if [existDbVal $thisNS $child _text] {
			cgi_li "[cgi_bold [getDbVal $thisNS $child title]] [cgi_nl]\
                                 [getDbVal $thisNS $child _text]"
		    } else {
			cgi_li "[cgi_bold [getDbVal $thisNS $child title]]"
		    }
		    cgi_br
		    cgi_br
		} else {
		    # Child should be a script
		    # Verify that child has active url and is part of 
		    # the nodeScriptList
		    if {[existDbVal $thisNS $child url] && \
			([lsearch -exact $nodeScriptList $child] != -1)} {
			if [existDbVal $thisNS $child _text] {
			    cgi_li "[cgi_link $child] [cgi_nl] [getDbVal $thisNS $child _text]"
			} else {
			    cgi_li "[cgi_link $child]"
			}
			cgi_br
			cgi_br
		    }
		}

		# Call make cascade recursively if child has children.
		if {[existDbVal $thisNS $child children] &&\
			[llength [getDbVal $thisNS $child children]] > 0} {
		    makeCascade $child
		}
	    }
	}
    }
    return 0
}

# CODEINFO BEGIN ######################################################

# info proc compChild> Procedure to sort children based on indices in
#     the master DB

# arg child1> Name of first child
# arg child2> Name of second child
# return> Child with lower database index

proc webSubmit::master::compChild {child1 child2} {
    variable DB
    variable thisNS

    set id1 [getDbVal $thisNS $child1 index]
    set id2 [getDbVal $thisNS $child2 index]
    if {$id1 < $id2} {
	return -1
    } elseif {$id1 > $id2} {
	return 1
    } else {
	return 0
    }
}


# CODEINFO BEGIN ######################################################

# info compAbs> Procedure to sort entries based on absolute indices in 
#     the master DB

# arg key1> Name of first key
# arg key2> Name of second key
# return> Child with lower database index

proc webSubmit::master::compAbs {key1 key2} {
    variable DB
    variable thisNS
    set id1 [getDbVal $thisNS $key1 auto]
    set id2 [getDbVal $thisNS $key2 auto]
    if {$id1 < $id2} {
	return -1
    } elseif {$id1 > $id2} {
	return 1
    } else {
	return 0
    }
}


# CODEINFO BEGIN ######################################################

# info defineLinks> Build up link information for master page

# arg dbNS> Namespace containing database
# arg cgiHead> CGI prefix for URL
# return keyList> List of dadabase keys of active scripts OR Error

# loc-sclr dbKeyList> List of keys for database
# loc-arr DB> Database array

proc webSubmit::master::defineLinks {dbNS cgiHead} {
    # Variable associations
    alias $dbNS DB dbKeyList
    variable thisNS

    set keyList ""
    # Build up link information for all ACTIVE scripts
    foreach key $dbKeyList {
	if ![existDbVal $thisNS $key] {
	    # Error - required variable nonexistent
            set errorMsg "ERROR (defineLinks): Key value \"$key\"\
                does not exist"
	    return -code error $errorMsg 
	}

    	foreach var [list moduleName type visible state build title\
            defaultMode host bqs] {
	    set $var [getDbVal $dbNS $key $var]
	}

	if {$type == "SCRIPT" && $visible == "ON" && $state == "ON"} {
	    if {$build == "WRAP"} {
		set url "$cgiHead/$formWrapper.cgi"
	    } else {
		set url "$cgiHead/$moduleName.cgi"
	    }
	    set args [queryString hostName $host moduleName $moduleName\
                 bqs $bqs mode $defaultMode title $title]
	    if {$args != ""} {
		set url "$url?$args"
	    }
	    setDbVal $dbNS "$url" $key url 
	    cgi_link $key $title $url
 	    lappend keyList "$key"
	}
    }

    return $keyList
}


























































































