# WebSubmit 2.0 wsNetwork.tcl @(#)wsNetwork.tcl	1.49 SCCS/s.wsNetwork.tcl  99/05/20 

# AUTHORS #############################################################

#  Authors: John Koontz, Ryan McCormack, Judith Devaney
#  Original Development Work: Robert Lipman, John Koontz, Katherine
#      Pagoaga
# WebSubmit 2.0 name @(#)wsUtil.tcl	1.64 SCCS/s.wsUtil.tcl  99/05/19 

# Copyright Information ###############################################

# This software was developed at the National Institute of Standards
# and Technology by employees of the Federal Government in the course
# of their official duties.  Pursuant to title 17 Section 105 of the
# United States Code this software is not subject to copyright
# protection and is in the public domain.

# CODEINFO BEGIN ######################################################

# info package wsNetwork.tcl > This is a package that contains service
#     routines for the WebSubmit user interface package.

# CODEINFO END ########################################################


# Package Declaration
package provide wsNetwork 1.0

# Namespace Definition
namespace eval webSubmit::network {
    # Procedures to be exported
    namespace export putFile
    namespace export getFile
    namespace export scpFile
    namespace export sshCmdExec
    namespace export getRemEnv
    namespace export remFileType

    # Variable declarations
    variable thisNS [namespace current]
    variable wsNS [namespace parent]
    variable upNS [namespace parent]
    variable debug 0

    # Procedures to be imported
    catch {namespace import ${wsNS}::util::*}
    alias $wsNS unixScp unixSsh unixFile
    variable scp $unixScp
    variable ssh $unixSsh
    variable file $unixFile

    # SSH Configuration information
    ;;readOnlyArray SSHCONFIG "arno.nist.gov,port" 22
    ;;readOnlyArray SSHCONFIG "amur.nist.gov,port" 22
    ;;readOnlyArray SSHCONFIG "danube.nist.gov,port" 22
    ;;readOnlyArray SSHCONFIG "rapid.nist.gov,port" 22
}

#######################
# Procedure definitions
#######################


# CODEINFO BEGIN
# info> Procedure to copy files between a client and server.  Currently
#  implemented using scp (Secure CoPy).  Capable of file transfer in either
#  direction based on "op" argument.

# arg op> Operation to perform (get = Copy from remote, put = Copy to remote)
# arg locFile> Local file name
# arg remHost> Remote hostname
# arg clobber> Overwrite file on target system if it exists (Overwrite=1)
# arg remUser> User name on remote system
# arg remFile> Filename on remote system
# arg args> Command options for scp
# return> Status (0 = Ok)
# loc-sclr op> Operation to perform (get = Copy from remote, put = Copy to remote)
# loc-sclr locFile> Local file name (relative or absolute)
# loc-sclr clobber> Overwrite file on target system if it exists (Overwrite=1)
# loc-sclr remHost>  Remote hostname
# loc-sclr remUser> User name on remote system
# loc-sclr remFile> Filename on remote system (relative or absolute)
# loc-sclr args> Command options for scp
proc webSubmit::network::scpFile  {locFile op clobber remUser remHost remFile args} {
    # Variable declarations
    variable scp
    variable debug
    variable SSHCONFIG

    if ![info exists SSHCONFIG($remHost,port)] {
	set port 22
    } else {
	set port $SSHCONFIG($remHost,port)
    }

    # Make sure sshd is active on remHost
    if ![portIsActive $remHost $port] {
	return -code error "$remHost does not appear to be listening on sshd port $port...Exiting"
    }

    # Translate filenames to absolute paths for clarity
    set home [getRemEnv HOME $remUser $remHost]
    set remFile [mkPathAbs $remFile $home]
    set locFile [mkPathAbs $locFile [pwd]]

    # Check to see if overwrite (clobber) specified
    if {!($clobber)} {
	# Check for existence of target file and exit if it already exists
	#  and is not a directory
	if {$op == "put"} {
	    if {[remFileExist $remUser $remHost $remFile] && \
		    ([remFileType $remUser $remHost $remFile] != "directory")} {
		return "WARNING: Remote target file exists and overwrite not specified - File not copied."
	    }
	} else {
	    if {[file exists $locFile] && ![file isdirectory $locFile]} {
		return "WARNING: Local target file exists and overwrite not specified - File not copied."
	    }
	}
    }

    # Compensate for null arguments to scp
    if {($args == "") || ($args == "{}")} {
	set SCP $scp
    } else {
	set SCP [join [list $scp $args]]
    }

    # Use scp to transfer file between client and server
    if {$op == "get"} {
	set cpCmd "webSubmit::util::execCgi \"$SCP -B -P $port ${remUser}@${remHost}:$remFile $locFile\""
	set msg "scpFile: Copied \"$remFile\" to \"$locFile\""
    } elseif {$op == "put"} {
	set cpCmd "webSubmit::util::execCgi \"$SCP -B -P $port $locFile ${remUser}@${remHost}:$remFile\""
	set msg "scpFile: Copied \"$locFile\" to \"$remFile\""
    } else {
	return -code error "File operation other than \"get\" or \"put\" specified"
    }

    if {$debug} {
	puts "\n\n"
	puts [webSubmit::util::header [list "SCP Information"] # -]
	puts "  SCP Command: \"$cpCmd\""
	puts "  Clobber: $clobber   Transfer direction: $op"
	puts "  User: $remUser   Host: $remHost   Port: $port"
	puts "  Local File: \"$locFile\""
	puts "  Remote File: \"$remFile\""
    }

    global errorCode
    if [catch {eval $cpCmd} result] {
	# Error detected during scp
	return -code error "$result (Code: $errorCode)"
    }
    return $msg
}


# CODEINFO BEGIN
# info> Procedure to execute commands on a remote host.  The routine supports
#  a variety of implementations.  The simplest UNIX utility is of course 'rsh'.
#  The more secure (recommended) version of this command is 'ssh'.  Each requires
#  that the remote system have the appropriate daemon running.

# arg remUser> Remote user name
# arg remHost> Remote hostname
# arg cmd> Command to be executed 
# arg outFile> Optional file to hold command output
# return> Result of command or Error
# loc-sclr remUser> Remote user name
# loc-sclr remHost> Remote hostname
# loc-sclr cmd> Command to be executed 
# loc-sclr outFile> Optional file to hold command output
proc webSubmit::network::sshCmdExec {remUser remHost cmd {outFile ""}} {
    # Variable declarations
    variable ssh
    variable debug
    variable SSHCONFIG

    if ![info exists SSHCONFIG($remHost,port)] {
	set port 22
    } else {
	set port $SSHCONFIG($remHost,port)
    }

    # Make sure sshd is active on remHost
    if ![portIsActive $remHost $port] {
	return -code error "$remHost does not appear to be listening on sshd port $port...Exiting"
    }

    # Wrap command inside delimiters for output parsing
    #  This avoids problems with users who have shell initialization scripts
    #  that do not check for non-interactive shells
    set boundary "---ws[pid]---"
    set cmd "echo $boundary;$cmd;echo $boundary"

    # Execute command via ssh - careful about interpretation of $cmd!!
    #   NOTE: Standard error redirected to stdout
    #   -q: Quiet mode
    #   -n: Do not attempt to read from stdin
    #   -p: Specifies port on which sshd listening
    #   -o "BatchMode = yes": Disables passphrase/password prompting
    if {$outFile != ""} {
	set command "$ssh -q -n -p $port -o \"BatchMode = yes\" -l $remUser $remHost \"$cmd\" >&@ $outFile"
    } else {
	set command "$ssh -q -n -p $port -o \"BatchMode = yes\" -l $remUser $remHost \"$cmd\""
    }
    
    if {$debug} {
	puts "\n\n"
	puts [webSubmit::util::header [list "SSH Command output:"] # -]
	puts "  User: $remUser   Host: $remHost   Port: $port"
	puts "  Command: \"$command\""
    }

    if [catch {webSubmit::util::execCgi "$command"} result] {
	# Error detected during command execution
	regsub -all -- "$boundary" $result "" result
	set result [string trim $result "\n"]
	return -code error $result
    } else {
	# Parse output to remove relevant data
	if [regexp -- "$boundary\n(.*)\n$boundary" $result match res1] {
	    return $res1
	} else {
	    return $result
	}
    }
}

# CODEINFO BEGIN
# info> Procedure to get an environment variable from a remote host
# arg varName> Name of environment variable to retrieve
# arg remUser> Remote user name
# arg remHost> Remote hostname
# return> Value of varName or NULL if it does not exist
# loc-sclr varName> Name of environment variable to retrieve
# loc-sclr remUser> Remote user name
# loc-sclr remHost> Remote hostname

proc webSubmit::network::getRemEnv {varName remUser remHost} {
    # Retrieve varName from user environment (if possible)
    set cmd "printenv $varName"
    catch {sshCmdExec $remUser $remHost "$cmd"} value
    
    # Check result of ssh output
    if [string match *ERROR* $value] {
	# Variable does not exist or error during exec
	return -code error "$value"
    } else {
	return $value
    }
}

# CODEINFO BEGIN
# info> Procedure to retrieve a small file from a remote system
# arg remUser> Remote user name
# arg remHost> Remote hostname
# arg remFile> Remote filename
# return> File or Error
# loc-sclr remUser> Remote user name
# loc-sclr remHost> Remote hostname
# loc-sclr remFile> Remote filename
proc webSubmit::network::getRemFile {remUser remHost remFile} {
    # Make sure file is text
    set type [remFileType $remUser $remHost $remFile]
    if {$type != "text" && $type != "script"} {
	return -code error "Unable to retrieve files of format \"$type\""
    }
    set tmp [sshCmdExec $remUser $remHost "cat '$remFile'"]
    if [string match *ERROR* $tmp] {
	# Error during retrieval of job file
	return -code error "$tmp"
    }
    set file [split $tmp "\n"]
    return $file
}

# CODEINFO BEGIN
# info> Procedure to determine remote file type
# arg remUser> Remote user name
# arg remHost> Remote hostname
# arg remFile> Remote filename
# return> File type
# loc-sclr remUser> Remote user name
# loc-sclr remHost> Remote hostname
# loc-sclr remFile> Remote filename
# loc-sclr type> Remote file type
proc webSubmit::network::remFileType {remUser remHost remFile {followLink 1}} {
    variable file

    # Verify file existence
    #  Unfortunately, the UNIX 'file' command does not return a non-zero
    #  exit status if the file in question does not exist
    if ![remFileExist $remUser $remHost $remFile] {
	return "null"
    }

    # Parse output from 'file' to determine type
    if [catch {webSubmit::network::sshCmdExec $remUser $remHost "$file '$remFile'"} tmp] {
	# Error
	return -code error "$tmp"
    } else {
	if [regexp {[^:]+: *(.*)} $tmp match type] {
	    switch -glob $tmp {
		*data* {set type "data"}
		*directory* {set type "directory"}
		*link* {set type "link"}
		*executable* {set type "executable"}
		*special* {set type "special"}
		*text* {set type "text"}
		*script* {set type "script"}
		default {set type "unknown"}
	    }
	} else {
	    set type "unknown"
	}

	# Recursively descend through symbolic links if desired to determine root
	#  file type
	if {$type == "link" && $followLink} {
	    set tmp [webSubmit::network::sshCmdExec $remUser $remHost "/bin/ls -l '$remFile'"]
	    regexp {.*-> *(.*)} $tmp match rootFile
	    set type [remFileType $remUser $remHost $rootFile]
	}
	return $type
    }
}

# CODEINFO BEGIN
# info> Procedure to check for existence of a file on a remote system
# arg remUser> Remote user name
# arg remHost> Remote hostname
# arg remFile> Remote filename
# return> 1 if file exists, 0 otherwise
# loc-sclr remUser> Remote user name
# loc-sclr remHost> Remote hostname
# loc-sclr remFile> Remote filename
proc webSubmit::network::remFileExist {remUser remHost remFile} {
    catch {sshCmdExec $remUser $remHost "/bin/ls -d '$remFile'"} result

    if ![string match "$remFile" $result] {
	# File does not exist
	set exist 0
    } else {
	set exist 1
    }
    return $exist
}

# CODEINFO BEGIN
# info> Procedure to open a pipe to an SSH process
# arg remUser> Remote user name
# arg remHost> Remote hostname
# arg cmd> Command to be executed 
# return> Result of command or Error
# loc-sclr remUser> Remote user name
# loc-sclr remHost> Remote hostname
# loc-sclr cmd> Command to be executed 
proc webSubmit::network::sshPipe {remUser remHost cmd} {
    # Variable declarations
    variable ssh
    variable done
    variable debug

    # Execute command via ssh - careful about interpretation of $cmd!!
    #   NOTE: Standard error redirected to stdout
    #   -q: Quiet mode
    #   -n: Do not attempt to read from stdin
    set pipe [open "|$ssh -n -q -o \"BatchMode = yes\" -l $remUser $remHost \"$cmd\" 2>@ stdout" r]

    if {$debug} {
	set command "|$ssh -n -q -o \"BatchMode = yes\" -l $remUser $remHost \"$cmd\" 2>@ stdout"
	puts "\n\n"
	puts [webSubmit::util::header [list "SSH Pipe:"] # -]
	puts "  User:\"$remUser\" Host:\"$remHost\""
	puts "  Command: \"$command\""
    }

    # Configure reader for pipe
    fileevent $pipe readable [list webSubmit::network::pipeReader $pipe]
    fconfigure $pipe -buffering line

    # Wait for completion of output from pipe
    vwait webSubmit::network::done
}


# CODEINFO BEGIN
# info> Procedure to read data from a pipe
# arg pipe> Name of pipe from which to read data
# return> None
# loc-sclr pipe> Name of pipe from which to read data
# wsNetwork-sclr done> Variable to indicate pipe closure
proc webSubmit::network::pipeReader {pipe} {
    variable done
    variable debug

    if [eof $pipe] {
	if {$debug} {
	    puts "  Closing pipe $pipe..."
	}
	catch {close $pipe}
	set done 1
	return
    }

    set getRet [gets $pipe line]
    if {$debug} {
	puts "  Reading data from pipe..."
	puts "  DATA: \"$line\""
	puts "  Return: $getRet"
    }

    if {$getRet < 0} {
	# Incomplete line"
    } else {
	# Process data coming from pipe
	puts "[cgi_sample $line]"
    }
}

# CODEINFO BEGIN
# info> Procedure to check whether a specified remote file is a writable
#  object.  If the remote file exists and is not a directory, the procedure
#  returns 0.  If the remote file exists and is a directory, or if the
#  remote file does not exist, the procedure returns 1.

# arg remUser> Remote user
# arg remHost> Remote hostname
# arg remFile> Remote filename
# return> 1 = Object writable, 0 otherwise
# loc-sclr remUser> Remote user
# loc-sclr remHost> Remote hostname
# loc-sclr remFile> Remote filename
proc webSubmit::network::remWritable {remUser remHost remFile} {
    set writable 0
    set remExist [webSubmit::network::remFileExist $remUser $remHost $remFile]
    if {$remExist} {
	set remType [webSubmit::network::remFileType $remUser $remHost $remFile]
	if {$remType == "directory"} {
	    set writable 1
	}
    } else {
	set writable 1
    }
    return $writable
}
	
# CODEINFO BEGIN
# info> Procedure to write a data set to a remote file
# arg remUser> Remote user
# arg remHost> Remote hostname
# arg data> Data to be written to file
# arg remFile> Remote filename
# arg clobber> Overwrite remote file if it exists
# return> Status
# loc-sclr remUser> Remote user
# loc-sclr remHost> Remote hostname
# loc-sclr data> Data to be written to file
# loc-sclr remFile> Remote filename
# loc-sclr clobber> Overwrite remote file if it exists
proc webSubmit::network::data2Remote {data remUser remHost remFile {clobber 0}} {
    variable ssh

    # Check for overwrite
    set writeAble [webSubmit::network::remWritable $remUser $remHost $remFile]
    if {$clobber} {
	if [webSubmit::network::remFileExist $remUser $remHost $remFile] {
	    # Remove remote file
	    if [catch {webSubmit::network::sshCmdExec $remUser $remHost "/bin/rm '$remFile'"} out] {
		return "Unable to delete remote file \"$remFile\" ($out)"
	    }
	}
    } else {
	if {!($writeAble)} {
	    return -code error "Overwrite not specified for existing remote file \"$remFile\""
	}
    }

    # Write data to remote file
    set dataLength [string length $data]
    if [catch {open "|$ssh -q -l $remUser $remHost \"cat > '$remFile'\"" w} pipeId] {
	return -code error "ERROR(data2remote): Could not write data to \"$remFile\" ($pipeId)"
    } else {
	fconfigure $pipeId -translation binary
        puts -nonewline $pipeId $data
	flush $pipeId
	close $pipeId
	return "$dataLength bytes written to \"$remFile\""
    }
}


# CODEINFO BEGIN ######################################################
# info proc portIsActive> Procedure to check whether a specified port
#     on a remote host is active (i.e., there is a process listening
#     on that port.

# arg host> Host name or IP address
# arg port> Port number
# return> 1 = Port active, 0 = No listener

proc webSubmit::network::portIsActive {host port} {
    if [catch {socket $host $port} fd] {
	return 0
    } else {
	close $fd
	return 1
    }
}







