# WebSubmit 2.0 wsUtil.tcl @(#)wsUtil.tcl	1.66 SCCS/s.wsUtil.tcl  99/07/12 

# AUTHORS #############################################################

#  Authors: John Koontz, Ryan McCormack, Judith Devaney
#  Original Development Work: Robert Lipman, John Koontz, Katherine
#      Pagoaga
# WebSubmit 2.0 name @(#)wsUtil.tcl	1.66 SCCS/s.wsUtil.tcl  99/07/12 

# Copyright Information ###############################################

# This software was developed at the National Institute of Standards
# and Technology by employees of the Federal Government in the course
# of their official duties.  Pursuant to title 17 Section 105 of the
# United States Code this software is not subject to copyright
# protection and is in the public domain.

# CODEINFO BEGIN ######################################################

# info package wsUtil > This is a package that contains service routines
#     for the WebSubmit user interface package.

# CODEINFO END ########################################################

# Package Declarations ################################################
package provide wsUtil 1.0

# Namespace Definition ################################################
namespace eval webSubmit::util {
    namespace export alias
    namespace export argParse
    namespace export ascii2hex
    namespace export bomb
    namespace export buildString
    namespace export checkChangeScope
    namespace export compSplit
    namespace export compareEnv
    namespace export createTags
    namespace export errLogMsg
    namespace export execCgi
    namespace export fileCurrent
    namespace export fileUsable
    namespace export getEnv
    namespace export getNSTree
    namespace export getUserName
    namespace export hash
    namespace export header
    namespace export initVars
    namespace export isUser
    namespace export jobLogMsg
    namespace export loadSerial
    namespace export mkPathAbs
    namespace export niceContinue
    namespace export nsExist
    namespace export nsJoin
    namespace export padString
    namespace export pipeRead
    namespace export qualifyNS
    namespace export readOnly
    namespace export readOnlyArray
    namespace export remLogMsg
    namespace export reqList
    namespace export safeArrSet
    namespace export safeLoadSerial
    namespace export safeSet
    namespace export serialize
    namespace export tclCat
    namespace export tclChmod
    namespace export tclDate
    namespace export tclGrep
    namespace export tclMkdir
    namespace export tclTouch
    namespace export trimSplit
    namespace export trimWhite
    namespace export undoChange
    namespace export writeLog

    # Variable definitions
    variable execError ""
    variable debug 0
    variable wsNS [namespace parent]
}

# Procedure definitions ###############################################

# CODEINFO BEGIN ######################################################
#

# info proc argParse> Generalized argument parser / tokenizer.  The parser
#     recognizes three types of tokens:
#   (1) Key-value pairs: '-key value'
#   (2) Switches: '--switchName'
#   (3) Regular arguments: Strings that don't fit either (1) or (2)

# arg args> Arguments to parse
# return> None

proc webSubmit::util::argParse {args} {
    uplevel "array set ARGV \"[list argList ""]\""
    upvar ARGV ARGV
    for {set i 0} {$i < [llength $args]} {incr i} {
        set this [lindex $args $i]
        set next [lindex $args [expr $i + 1]]
        if [regexp {^--(.*)} $this match key] {
	    # $this corresponds to a switch
	    set ARGV($key) 1
	} elseif [regexp {^-([^-]*)} $this match key] {
	    if ![string match -* $next] {
		set ARGV($key) $next
		incr i
	    } else {
		set ARGV($key) ""
	    }
        } else {
	    # Argument should be added to argList
	    lappend ARGV(argList) $this
        }
    }
} 

# CODEINFO BEGIN ######################################################
#

# info proc nsJoin> Procedure to join a group of strings into a
#     namespace name (either fully- or partially-qualified)

# arg args> [-partial] NSnames.  Specifies whether qualification is partial and
#     then list of names to be joined

# return> Fully- or partially-qualified namespace name corresponding to
#     the concatenation of $NSnames with "::" between each element

proc webSubmit::util::nsJoin {args} {
    if {[lindex $args 0] == "-partial"} {
	set ns [lindex $args 1]
	set compList [lreplace $args 0 1]
    } else {
	set compList $args
	set ns ""
    }

    foreach elt $compList {
	if {$elt != ""} {
	    append ns "::" [string trim $elt ::]
	}
    }
    return $ns
}


# CODEINFO BEGIN ######################################################

# info proc qualifyNS> Procedure to resolve a namespace name into
#     one that is fully-qualified.  If the name is already fully-qualified, 
#     nothing is done.  If partially-qualified, then the name for the
#     calling scope is prepended.  The syntax is as follows:
#
# qualifyNS scope NSname
#
#  Scope:
#    wsRoot: Name is qualified wrt webSubmit root namespace
#    global: Name is qualified wrt global namespace (::)
#    someNamespaceName: Name is qualified wrt specified name

# arg scope> Scope in which qualification will be made
# arg NS> Namespace name to qualify
# return> Fully-qualified namespace

proc webSubmit::util::qualifyNS {scope NS} {
    variable wsNS
    variable debug
    if [regexp {^::.*} $NS] {
	# Namespace is already fully-qualified
	return $NS
    } else {
	set nsHead ""
	switch -exact $scope {
	    wsRoot {set nsHead $wsNS}
	    global {set nsHead ""}
            defScope {set nsHead ""}
	    default {set nsHead $scope} 
	}

	if {$debug} {
	    puts "\nqualifyNS: Qualifiying namespace: scope header ($rootScope) = $nsHead"
	    puts "qualifyNS: Namespace to resolve: $NS"
	}

	if {$nsHead == "::"} {
	    set nsHead ""
	}
	return [nsJoin $nsHead $NS]
    }
}

# CODEINFO BEGIN ######################################################

# info proc alias>  Procedure to declare a group of variables within a 
#     namespace and to associate these variables with variables of the 
#     same name in the calling scope.  This is equivalent to a global 
#     statement, with the global namespace being replaced by the namespace 
#     specified in the argument.

# arg scope> -wsRoot, -global, -user=?, or -call.  See webSubmit::util::qualify
# arg theNS> Name of the relevant namespace
# arg args> List of variable names to be declared
# return> Status (Error = 1)
# loc-sclr fqNS> Fully-qualified namespace name corresp. to $theNS

proc webSubmit::util::alias {args} {
    variable debug

    # Parse argument list
    set ARGV(scope) defScope
    eval argParse $args
    set NS [lindex $ARGV(argList) 0]
    set vars [lreplace $ARGV(argList) 0 0]

    if {$debug} {
	puts "\nalias: Create alias for \"$vars\" contained in namespace \"$NS\""
	puts "alias: Alias being created in [uplevel "namespace current"]"
    }

    if ![nsExist -scope $ARGV(scope) $NS] {
	# Namespace does not exist - association not possible
	set errorMsg "ERROR (alias): Association failed - \"$NS\" does not exist"
	return -code error $errorMsg
    }

    # Qualify the namespace name with respect to the
    #  specified scope
    set fqNS [qualifyNS $ARGV(scope) $NS]

    foreach var $vars {
	# Make sure variable is declared in spaceName
	if {![info exists ${fqNS}::$var]} {
	    variable ${fqNS}::$var
	}

	# Create associated variable in the calling scope
	catch {uplevel "upvar 0 ${fqNS}::$var $var"}

    }
    return 0
}


# CODEINFO BEGIN ######################################################

# info proc nsExist> Procedure to check to see if a namespace is defined

# arg args> Argument list (passed on to qualifyNS)
# return> Status (Error = 1)

proc webSubmit::util::nsExist {args} {
    variable debug
    if {$debug} {
        puts "\nnsExist: Check existence of $theNS"
    }

    # Parse arguments - set scope to default if none given
    set ARGV(scope) defScope
    eval argParse $args

    # Qualify namespace name
    set fqNS [qualifyNS $ARGV(scope) $ARGV(argList)]

    if [catch {namespace parent $fqNS} out] {
	# Namespace does not exist
	return 0
    } else {
	return 1
    }
}

# CODEINFO BEGIN ######################################################

# info proc bomb> Procedure to handle bombs gracefully and send output 
#     through CGI

# arg proc> Name of procedure in which bomb occurred
# arg string> String to return in bomb error message
# return> None

# Superceded by cgiBomb

proc webSubmit::util::bomb {proc string} {
    # HTML Body
    cgi_http_head
    cgi_html {
	cgi_body link=antiquewhite vlink=rosybrown bgcolor=antiquewhite {
	    cgi_br
	    cgi_h2 "[cgi_bold [cgi_font size=+2 color=firebrick "WebSubmit Error:"]]"
	    cgi_hr
	    cgi_p "The following error message was generated by WebSubmit in procedure\
		    [cgi_bold $proc]:"
	    cgi_p "$string"
	}
    }
    exit
}


# CODEINFO BEGIN ######################################################

# info proc execCgi> Procedure to trap bogus errors generated by exec 
#     when performed in a CGI context.  Error is stripped from output of 
#     exec and returned, unless a real error is generated

# arg cmd > Command to evaluate in exec
# return > Output from exec if normal, or 'ERROR' if real error detected

# loc-sclr rVal > Output from exec if no error detected
# loc-sclr cmd >  Command to evaluate in exec
# wsUtil-sclr execError > Bogus error to trap (related to HTTP)

proc webSubmit::util::execCgi {cmd} {
    variable execError
    if ![info exists execError] {
	catch {exec echo} tmp
	set execError [string trim $tmp]
    }

    # Perform exec and catch error
    if [catch {eval exec $cmd} result] {
	# Check to see whether error contains bogus execError string
	if {($execError != "") && [string match "*$execError" "$result"]} {
	    # Error should be a false positive due to a CGI exec error
            # Assume everything but execError text is a message
	    set rVal [string trimright $result $execError]
	} else {
	    # Error has to be real
	    set rVal $result
	    global errorCode
	    if {$errorCode != "NONE"} {
		return -code error $rVal
	    }
	}
    } else {
	# No error detected - Return output of exec
	set rVal "$result"
    }
    return $rVal
}

# CODEINFO BEGIN ######################################################

# info proc fileCurrent>  Determine whether thatFile is current with 
#     respect to thisFile.  If thatFile does not exist, it is 
#     not current.  ThisFile must exist.

# arg-sclr thisFile> The first file involved in the comparison.
# arg-sclr thatFile> The second file involved in the comparison.
# return>  
#     False (0) if thisFile exists and is newer than thatFile, 
#     or thatFile does not exist.  
#     True (1) if thisFile exists and is older than thatFile.  
#     Exception if thisFile does not exist.

proc webSubmit::util::fileCurrent {thisFile thatFile} {
    if {![fileUsable $thisFile]} {
	set errMess "ERROR(fileNewer):  The primary file \"$thisFile\"\
            does not exist."
	return -code error $errMess
    } elseif {![fileUsable $thatFile]} {
	return 0
    } elseif {[file mtime $thisFile] < [file mtime $thatFile]} {
        return 1
    } else {
	return 0
    }
}

# CODEINFO BEGIN ######################################################

# info proc fileUsable>  Determine whether a file exists and is readable.

# arg-sclr theFile> The name of the file to check.
# return>  True (1) if the file exists and is readable, false (0) 
#     otherwise.

proc webSubmit::util::fileUsable {theFile} {
    if {![file exists $theFile] || ![file readable $theFile]} {
	return 0
    } else {
	return 1
    }
}

# CODEINFO BEGIN ######################################################

# info proc reqList> Procedure to check whether one list possesses at 
#     least the elements contained in another list

# arg varList> List of existing elements
# arg reqList> List of required elements
# return> 1 - List contains required elements, 0 - otherwise

proc webSubmit::util::reqList {varList reqList} {
    foreach elt $reqList {
	if {[lsearch -exact $varList $elt] == -1} {
	    return 0
	}
    }
    return 1
}


# CODEINFO BEGIN ######################################################

# info proc tclGrep> Procedure to perform the equivalent of UNIX 'grep'
#     on a file.  The purpose of writing such a routine is to provide for 
#     platform independence.

# arg file > File to search
# arg string > String to search for in file
# return > List of lines containing string (Empty if no match), Error

proc webSubmit::util::tclGrep {string file} {    
    # Open file for reading
    if [catch {open $file r} fileId] {
	return -code error $fileId
    }

    # Scan for string in file
    set matchList ""
    while {[gets $fileId line] >= 0} {
	if [regexp "$string" $line match] {
	    lappend matchList $line
	}
    }
    close $fileId

    if {$matchList != ""} {
	return $matchList
    } else {
	# Null match
	return -code error
    }
}

	
# CODEINFO BEGIN ######################################################

# info proc tclTouch> Procedure to create a zero-size file.  Equivalent 
#     to the UNIX 'touch' command.  Provided for platform independence.

# arg fileName > Name of file to create (either absolute or relative path)
# return > Error Status

proc webSubmit::util::tclTouch {fileName} {
    if [catch {open $fileName w} id] {
	return -code error $id
    }
    close $id
    return 0
}


# CODEINFO BEGIN ######################################################

# info proc trimSplit> Procedure to split a string into a list and remove 
#     any whitespace around entries in the list.  This is to deal with 
#     situations due to spaces in a list with non-whitespace separators.
#     In addition, and separator characters char are ignored if they fall
#     within quotation marks (").

# arg-scalar string > String to be split into list.
# arg-scalar char > Separator.
# return > List corresponding to string with internal white space removed

proc webSubmit::util::trimSplit {string char} {
    # Split into list containing possible white-space
    set tmpList [split $string $char]

    if {$char != "\""} {
	set i 0 
	while {$i < [expr [llength $tmpList] - 1]} {
	    set quoteCount [regsub -all \" [lindex $tmpList $i] {} ignore]
	    if {[expr $quoteCount % 2 == 1]} {
		set tmpList [lreplace $tmpList $i [expr $i + 1] \
		    "[lindex $tmpList $i]$char[lindex $tmpList [expr $i + 1]]"]
	    } else {
		incr i
	    }
	}
    }

    set compList ""
    foreach elt $tmpList {
        lappend compList [string trim $elt]
    }
    return $compList
}
    

# CODEINFO BEGIN ######################################################

# info proc compSplit> Procedure to split a string into a list and remove 
#     all empty entries from the list.  This is to deal with situations
#     where the string that is being split contains sections of 
#     white-space that need to be compressed

# arg string > String to be split into list
# return > List corresponding to string with internal white space removed

proc webSubmit::util::compSplit {string} {
    # Split into list containing possible white-space
    set tmpList [split $string]

    set compList ""
    foreach elt $tmpList {
	if {$elt != ""} {
	    # Add to list
	    lappend compList $elt
	}
    }
    return $compList
}
    

# CODEINFO BEGIN ######################################################

# info proc getEnv> Procedure to find and return the value of a specified
#     environment variable (global).  If the variable does not exist, an
#     error is raised

# arg varName> Name of the environment variable to check
# return> Value of varaibel or error

# glob-arr env> Environment variable array

proc webSubmit::util::getEnv {varName} {
    global env
    if [info exists env($varName)] {
	return $env($varName)
    } else {
	set errorMsg "ERROR (getEnv): \"$varName\" is not a valid\
            environment variable"
	return -code error $errorMsg 
    }
}


# CODEINFO BEGIN ######################################################

# info proc compareEnv> Procedure to compare the value of an environment
#     variable with a specified value.  This can be used, for example, 
#     to determine if the REMOTE_HOST IP address corresponds to the 
#     WebSubmit server.

# arg varName> Environment variable to compare
# arg envValue> Value to compare against
# return> 1 = Equal, 0 = Unequal, Error if undefined

proc webSubmit::util::compareEnv {varName envValue} {
    if [catch {getEnv $varName} result] {
	# Invalid environment
	return -code error $result
    } elseif {$value == $result} {
	# Values match
	return 1
    }
    return 0
}
	

# CODEINFO BEGIN ######################################################

# info proc writeLog> Procedure to write a message to a WebSubmit log
#     file.  The procedure provides for multiple log files and an
#     arbitrary message.  CURRENT STATUS: Unused

# arg logFile> Name of the logfile to be written (Absolute or relative path)
# arg msgList> List of messages to be written to logFile
# return> Error status

proc webSubmit::util::writeLog {logFile msg} {
    # Open logFile and append message
    if [catch {open $logFile a} logId] {
        set errorMsg "ERROR (writeLog): $logId"
	return -code error $errorMsg
    }

    # Write message to log with current date prepended
    puts $logId "\[[tclDate]\]: $msg"
    close $logId
}


# CODEINFO BEGIN ######################################################

# info proc remLogMsg> Procedure to build up an access log message.  
#     This message is the one printed in the log files when the main 
#     WebSubmit page is accessed from a remote location (i.e., a URL 
#     outside the WS hierarchy.  CURRENT STATUS: Unused
   
# arg > NONE
# return> Log message

proc webSubmit::util::remLogMsg {} {
    global env
    set varList [list REMOTE_ADDR]
    
    # Make sure proper CGI environment exists
    foreach var $varList {
	if ![info exists env($var)] {
	    # Invalid CGI environment
            set errorMsg "Invalid CGI environment for $var"
	    return -code error $errorMsg
	}
    }

    # Build log message
    lappend msg "Access to WebSubmit server from $env(REMOTE_ADDR)"
    return $msg
}


# CODEINFO BEGIN ######################################################

# info proc errLogMsg> Procedure to build up an error log message.  
#     This message is the one printed in the log files when a WebSubmit 
#     error occurs.  CURRENT STATUS: Unused
   
# arg status> Error status code
# arg module> Module from which error was generated
# arg string> Text string to print with error message
# return> Log message

proc webSubmit::util::errLogMsg {status module string} {
    global env
    set varList [list REMOTE_HOST REMOTE_ADDR REMOTE_USER]
    
    # Make sure proper CGI environment exists
    foreach var $varList {
	if ![info exists env($var)] {
	    # Invalid CGI environment
	    set errorMsg "Invalid CGI environment for $var"
	    return -code error $errorMsg
	}
    }

    # Build log message
    lappend msg "$module ($status): $string"
    lappend msg "      ($env(REMOTE_USER)@$env(REMOTE_HOST)\
        ($env(REMOTE_ADDR)))"
    return $msg

}


# CODEINFO BEGIN ######################################################

# info proc jobLogMsg> Procedure to build up job submission log message.
#     This message is the one printed in the log files when jobs are
#     submitted via WebSubmit. CURRENT STATUS: Unused
   
# arg userId> WebSubmit userId
# arg target> Target machine to which job was submitted
# arg module> Name of module used to submit job
# return> Log message

proc webSubmit::util::jobLogMsg {userId target module} {
    # Build log message
    set msg "User=$userId Target=$target Module=$module"
    return [list $msg]
}


# CODEINFO BEGIN ######################################################

# info proc mkPathAbs> Procedure to make a relative pathname absolute

# arg path> Path to fix
# arg root> Root to prepend to path if it is relative
# return> Absolute path corresponding to input

proc webSubmit::util::mkPathAbs {path root} {
    # Strip trailing slash from root path
    set root [string trimright $root /]

    # Verify that root is in fact an absolute head
    if ![regexp {^/.*} $root match] {
	return $path
    }

    # Check path
    if {[file pathtype $path] == "relative"} {
	set path [join [list $root $path] /]
    }
    return $path
}


# CODEINFO BEGIN ######################################################

# info proc buildString> Procedure to build a string of repeated characters

# arg len> Length of string to create
# arg char> Character to use in string
# return> String of length 'len' containing only 'char'

# loc-sclr string> String to return

proc webSubmit::util::buildString {len char} {
    set string $char
    if {$len > 1} {
	for {set i 1} {$i < $len} {incr i} {
	    append string $char
	}
    }
    return $string
}


# CODEINFO BEGIN ######################################################

# info proc header> Procedure to build a simple text header for output
#     (debugging)

# arg stringList> String to enclose in header
# arg topchar> Character across top of header
# arg botchar> Character across bottom of header
# return> Header text as string value

proc webSubmit::util::header {stringList topchar botchar} {
    set max 0
    # Find maximum length string in stringList
    foreach string $stringList {
	set length [string length $string]
	if {$length > $max} {
	    set max $length
	}
    }

    # Retrieve top and bottom strings
    set top [buildString $max $topchar]
    set bot [buildString $max $botchar]
    set head "\n$top"
    foreach string $stringList {
	append head "\n$string"
    }
    append head "\n$bot"
    return $head
}


# CODEINFO BEGIN ######################################################

# info proc padString> Procedure to pad a string on the left and right 
#     with blanks

# arg str> String to pad
# arg char> Padding character
# arg totLen> Length to pad to
# return> The padded string

proc webSubmit::util::padString {str char totLen} {
    set len [string length $str]
    if {$totLen <= $len} {
	return $str
    }

    set pad [expr ($totLen - $len)]
    set numL [expr $pad / 2]
    if {[expr $pad % 2] == "0"} {
	set numR $numL
    } else {
	set numR [expr $numL + 1]
    }

    set padL ""
    set padR ""
    foreach elt [list R L] {
	for {set i 1} {$i <= [set num${elt}]} {incr i} {
	    append pad${elt} "$char"
	}
    }
    append padL $str $padR
    return $padL
}

# CODEINFO BEGIN ######################################################

# info proc initVars> Procedure to initialize a list of scalars in the 
#     *calling scope* to empty strings.  If the variable already exists,
#     its value is not changed.

# arg varList> List of variables to initialize
# return> None

proc webSubmit::util::initVars {varList} {
    foreach var $varList {
	if ![uplevel "info exists $var"] {
	    uplevel "set $var \"\""
	}
    }
    return
}

# CODEINFO BEGIN ######################################################

# info proc serialize> Procedure to write a collection of data to a Tcl
#    file.  In this fashion, the data structure can be reproduced simply
#    by source-ing the file.  In an attempt to preserve list structures,
#    it is assumed that any variable that has a name of the form "*List"
#    is in fact a list.  Arrays are written element by element; strings
#    are simply 'set'.

# arg varList> List of variable *names* to serialize
# arg ns> Namespace (absolute) in which variables exist
# arg file> Name of file to create
# return> None

proc webSubmit::util::serialize {ns varList file} {
    # Create aliases to variables in ns
    eval alias $ns $varList

    # Check for existence of parent directory
    set dir [file dirname $file]
    if ![file exists $dir] {
	file mkdir $dir
	tclChmod 775 $dir
    }

    # Open file for serialization
    if [catch {open $file w} sId] {
	return -code error "ERROR(serialize): $sId"
    }

    # Write variables to file in Tcl source-able format
    puts $sId "# Serialization of data from namespace $ns"
    puts $sId "#  Date: [tclDate]"
    puts $sId "namespace eval $ns \{"
    foreach var $varList {
	if [array exists $var] {
	    # Variable is an array
	    foreach key [lsort [array names $var]] {
		puts $sId "   set ${var}($key) \{[set ${var}($key)]\}"
	    } 
	} elseif {[info exists $var] && [regexp {.*List$} $var]} {
	    # Variable is a list
	    foreach elt [set $var] {
		puts $sId "   lappend $var \{$elt\}"
	    }
	} elseif [info exists $var] {
	    # Variable is a string
	    puts $sId "   set $var \{[set $var]\}"
	}
    }
    puts $sId "\}\nreturn 0"
    close $sId

    # Change file modes on serialized file to make it world-readable
    tclChmod 664 $file

    return
}

# CODEINFO BEGIN ######################################################

# info proc loadSerial> Procedure to source a serialized version of a
#     collection of data.  

# arg fileName> Name of the file to source
# return> 0 = Success, 1 = Error

proc webSubmit::util::loadSerial {fileName} {
    # Verify that source file exists and is a regular file
    if [file isfile $fileName] {
	if [catch {source $fileName} err] {
	    set ret 1
	} else {
	    set ret 0
	}
    } else {
	set ret 1
    }
    return $ret
}



# CODEINFO BEGIN ######################################################

# info proc readOnly> Procedure to store read-only variable values
#     and to place variable traces on these variables

# arg varName> Name of the variable to protect
# arg value> Value of variable
# arg scope> Scope in which variable is to be defined
# return> None

proc webSubmit::util::readOnly {varName value {scope ""}} {
    variable wsNS

    # Set up namespace for variable. Default = Namespace of caller
    if {$scope == ""} {
	set scope [uplevel "namespace current"]
    }
    set tmpVar [namespace tail $varName]
    set var "${scope}::$tmpVar"
    
    # Create variable if it does not already exist
    if ![info exists $var] {
	variable $var "$value"
    } else {
	set $var "$value"
    }
    
    # Store original value of variable
    if ![info exists ${wsNS}::VARSAFE($var)] {
	set ${wsNS}::VARSAFE($var) "$value"
    }
    
    # Create variable trace if no traces exist
    if {[trace vinfo $var] == ""} {
	trace variable $var wu [list webSubmit::util::undoChange $var]
    }
}

# CODEINFO BEGIN ######################################################

# info proc readOnlyArray> Procedure to store read-only array variable values
#     and to place variable traces on these variables

# arg arrName> Name of the array variable to protect
# arg index> Array index
# arg value> Value of variable
# arg scope> Scope in which variable is to be defined
# return> None

proc webSubmit::util::readOnlyArray {arrName index value {scope ""}} {
    variable wsNS

    # Set up namespace for variable. Default = Namespace of caller
    if {$scope == ""} {
	set scope [uplevel "namespace current"]
    }
    set tmpVar [namespace tail $arrName]
    set arr "${scope}::$tmpVar"
    
    # Create variable if it does not already exist
    if ![info exists $arr] {
	variable ${arr}() {}
	set ${arr}($index) "$value"
    } elseif ![info exists ${arr}($index)] {
	set ${arr}($index) "$value"
    }
    
    # Append to list of read-only arrays
    if {[lsearch -exact [set ${wsNS}::VARSAFE(ArrayList)] $arr] == -1} {
	lappend ${wsNS}::VARSAFE(ArrayList) $arr
    }

    # Store original value of variable
    if ![info exists ${wsNS}::VARSAFE($arr,$index)] {
	set ${wsNS}::VARSAFE($arr,$index) "$value"
    }
    
    # Create variable trace if no traces exist
    if {[trace vinfo $arr] == ""} {
	trace variable $arr wu [list webSubmit::util::undoArrChange $arr]
    }
}

# CODEINFO BEGIN ######################################################

# info proc undoChange> Implements a read-only variable scheme.  Any time
#     a variable's value is written or unset, it is reset to its original
#     value as stored in webSubmit::VARSAFE

# arg varName> Name of variable
# arg callName> Name of variable used in call 
# arg index> Index if varName is an array name
# arg op> Operation occurring on variable (r=read,w=write,u=unset)
# return> None

proc webSubmit::util::undoChange {varName callName index op} {
    variable wsNS
    upvar 0 ${wsNS}::VARSAFE VARSAFE
    upvar $varName var

    if [info exists VARSAFE($varName)] {
	# Retrieve original value from safe
	set origValue $VARSAFE($varName)

	# Switch on op and restore value
	switch $op {
	    w {set var $origValue}
	    u {set var $origValue
		trace variable $varName wu [list webSubmit::util::undoChange $varName]}
	}
    } else {
	# Store new read-only value
	if {$op == "w"} {
	    set value "$var"
	    set VARSAFE($varName) $value
	}
    }
}


# CODEINFO BEGIN ######################################################

# info proc undoArrChange> Implements a read-only array scheme.  Any time
#     a variable's value is written or unset, it is reset to its original
#     value as stored in webSubmit::VARSAFE

# arg arrName> Name of variable
# arg callName> Name of variable used in call 
# arg index> Index if varName is an array name
# arg op> Operation occurring on variable (r=read,w=write,u=unset)
# return> None

proc webSubmit::util::undoArrChange {arrName callName index op} {
    variable wsNS
    upvar 0 ${wsNS}::VARSAFE VARSAFE
    upvar 0 $arrName arr

    # Search for array in VARSAFE
    if {[lsearch -exact $VARSAFE(ArrayList) $arrName] == -1} {
	# Variable is not array
	return -code error "ERROR(undoArrChange): Variable \"$arrName\" is not array"
    } else {
	# Get list of keys that need to be restored
	switch $op {
	    w {set keyList "$arrName,$index"}
	    u {
		if {$index == ""} {
		    # Restore entire array
		    set keyList [array names VARSAFE $arrName,*]
		} else {
		    set keyList "$arrName,$index"
		}
	    }
	}
    }

    foreach key $keyList {
	if [info exists VARSAFE($key)] {
	    # Retrieve original value from safe
	    set origValue $VARSAFE($key)
	    
	    # Extract index to restore
	    regexp {[^,]+,(.*)} $key match index

	    # Switch on op and restore value
	    switch $op {
		w {set arr($index) $origValue}
		u {set arr($index) $origValue}
	    }
	} else {
	    # Store new read-only value
	    if {$op == "w"} {
		set VARSAFE($key) "$arr($index)"
	    }
	}
    }

    # Restore trace if entire array was unset
    if {$op == "u"} {
	if {[trace vinfo $arrName] == ""} {
	    trace variable $arrName wu [list webSubmit::util::undoArrChange $arrName]
	}
    }
}


# CODEINFO BEGIN ######################################################

# info proc checkChangeScope> Trace procedure to determine the environment
#      from which a variable change was invoked.  If invoked from a
#      valid environment, no action is taken.  If invoked from an
#      insecure location (i.e., the global environment), then the
#      trace generates a securityBomb and exits

# arg varName> Name of variable
# arg index> Index if varName is an array name
# arg op> Operation occurring on variable (r=read,w=write,u=unset)
# return> None

proc webSubmit::util::checkChangeScope {varName index op} {
    set scope [uplevel "namespace current"]
    if ![string match *webSubmit* $scope] {
	# Procedure invoked from outside the webSubmit scope
	#  Drop a securityBomb
	set errString "SECURITY VIOLATION: Unauthorized variable access from scope \"$scope\""
	puts "$errString (Variable = $index)"
    }
}

# CODEINFO BEGIN ##########################################################
# info proc ascii2hex> Procedure to convert an ASCII character into a hex
#     representation suitable for a URL.  Optional argument is a kludge to
#     handle newline replacements

# arg char> Character to convert
# return> Hex value of character, preceded by '%'
proc webSubmit::util::ascii2hex {{char ""}} {
    if {$char != ""} {
	scan $char %c result
	return %[format %02X $result]
    } else {
	return "%0A"
    }
}


# CODEINFO BEGIN ##########################################################
# info proc tclChmod> Procedure to change file permissions (interface to 
#     UNIX chmod)
# arg mode> Mode for file(s)
# arg fileList> List of files to change modes
# return> Status

proc webSubmit::util::tclChmod {mode fileList} {
    variable wsNS
    set chmod [set ${wsNS}::unixChmod]
    if [catch {exec $chmod -R $mode $fileList} result] {
	return -code error $result
    }
    return 0
}


# CODEINFO BEGIN ##########################################################
# info proc tclMkdir>  Procedure to make directories 
# arg mode> Mode for directories
# arg dirList> List of pathnames of directories to make
# return> Status

proc webSubmit::util::tclMkdir {mode dirList} {
    foreach d $dirList {
	set subdirList [file split $d]
	set pathName ""
	foreach s $subdirList {
	    if {"$pathName" == ""} {
		set pathName $s
	    } else {
		set pathName [file join $pathName $s]
	    }
	    if ![file isdirectory $pathName] {
		file mkdir $pathName
		tclChmod $mode $pathName
	    }
	}
	if ![file isdirectory $d] {
	    set errMess "The pathName \"$d\" is not a directory."
	    return -code error $errMess
	}
	    
    } 
    return 0
}

# CODEINFO BEGIN ##########################################################
# info proc trimWhite> Procedure to trim white space from a list of variables
# arg varList> List of variables from which white space should be trimmed
# return> None

proc webSubmit::util::trimWhite {varList} {
    foreach var $varList {
	set value [uplevel "set $var"]
	set newValue [string trim $value]
	uplevel "set $var \{$newValue\}"
    }
    return 0
}


# CODEINFO BEGIN ############################################
# info proc pipeRead> A general fileevent handler for reading data
#     from a specified pipe and then closing the pipe once eof is
#     reached.  No special data processing is performed.  Data
#     from the pipe is simply returned to stdout.

# arg pipe> File descriptor for the pipe being read
# arg waitVarName> Variable being waited on for pipe close
# return> None

proc webSubmit::util::pipeRead {pipe {waitVarName pipeDone}} {
    upvar #0 $waitVarName done
    variable debug

    if [eof $pipe] {
	if {$debug} {
	    puts "  Closing pipe $pipe..."
	}
	catch {close $pipe}
	set done 1
	return
    }

    set getRet [gets $pipe line]
    if {$debug} {
	puts "  Reading data from pipe..."
	puts "  DATA: \"$line\""
	puts "  Return: $getRet"
    }

    if {$getRet < 0} {
	# Incomplete line
    } else {
	# Process data coming from pipe
	puts "$line"
    }
}


# CODEINFO BEGIN ###############################################
# info proc safeSet> Procedure transfer variables from a safe
#     interpreter into the master interpreter

# arg NS> Namespace context in which variables should be set
# arg varName> Name of variable to set 
# arg value> Value of variable

proc webSubmit::util::safeSet {varName value {NS ::}} {
    set ${NS}::varName $varName
    set ${NS}::value $value
    namespace eval $NS {
	if {[namespace current] != "::"} {
	    if [catch {set $varName $value} error] {
		puts "safeSet NS error: $error"
	    }
	} else {
	    if [catch {set $varName $value} error] {
		puts "safeSet global error: $error"
	    }
	}
    }
    return
}

# CODEINFO BEGIN ###############################################
# info proc safeArrSet> Procedure transfer array variables from a safe
#     interpreter into the master interpreter

# arg NS> Namespace context in which variables should be set
# arg arrName> Name of variable to set 
# arg arrDataList> List fo array data (alternating key-value pairs)

proc webSubmit::util::safeArrSet {arrName arrDataList {NS ::}} {
    set ${NS}::arrName $arrName
    set ${NS}::arrDataList $arrDataList
    namespace eval $NS {
	if {[namespace current] != "::"} {
	    if [catch {array set $arrName $arrDataList} error] {
		puts "safeArrSet NS error: $error"
	    }
	} else {
	    if [catch {array set $arrName $arrDataList} error] {
		puts "safeArrSet global error: $error"
	    }
	}
    }
    return
}


# CODEINFO BEGIN ##################################################
# info proc getNSTree> Procedure to retrieve the tree containing
#     all namespace names that are children (or their descendants)
#     within a given root namespace

proc webSubmit::util::getNSTree {{rootNS ::}} {
    variable nsList

    if [nsExist $rootNS] {
	# Retrieve children at the level of the tree and descend if necessary
	set children [namespace children $rootNS]
	if {$children == ""} {
	    return
	} else {
	    foreach child $children {
		lappend nsList $child
		getNSTree $child
	    }
	}
	return $nsList
    } else {
	return ""
    }
}


# CODEINFO BEGIN ##################################################
# info proc safeLoadSerial> Procedure to load serialized data via
#     a safe interpreter.

# arg fileName> Name of file to load
# return> Success = 1, Failure =0
proc webSubmit::util::safeLoadSerial {fileName} {
    variable wsNS
    alias $wsNS safeInterp

    # Verify that source file exists and is a regular file
    #   Source inside safe interpreter
    if [file isfile $fileName] {
	if [catch {interp invokehidden $safeInterp source $fileName} err] {
	    set ret 1
	} else {
	    set ret 0
	}
    } else {
	set ret 1
    }

    # Transfer variables from safe interpreter
    #  Do not transfer data from ::webSubmit or ::webSubmit::util namespaces
    #
    set nsList [$safeInterp eval getNSTree]
    foreach ns $nsList {
	set varList [$safeInterp eval info vars ${ns}::*]
	if {$varList != ""} {
	    foreach var $varList {
		set varName [namespace tail $var]
		if [$safeInterp eval array exists $var] {
		    set arrDataList [$safeInterp eval array get $var]
		    $safeInterp eval safeArrSet [list $varName $arrDataList $ns]
		} else {
		    set value [$safeInterp eval set $var]
		    $safeInterp eval safeSet [list "$varName" "$value" $ns]
		}
	    }
	}
    }

    # Delete all data in slave interpreter
    foreach ns $nsList {
	# Delete namespace
	$safeInterp eval namespace forget $ns
    }

    return $ret
}

# CODEINFO BEGIN #######################################################
# info proc hash> Procedure to return the MD5 hash of an input string.
#     The MD5 hash is a 128-bit hash, represented by 32 hex digits.

# arg string> String to hash
# return> MD5 hash of string

proc webSubmit::util::hash {string} {
    catch {unset d o}

    # Create necessary binary data objects
    #   Buffer = n-byte Input buffer (n = Input string length)
    #   Digest = 16-byte Output buffer for MD5 digest
    bin_new o buffer [string length $string]
    bin_new d digest 16
    set o(_str_) $string

    # Perform hash and return value
    md5 d o
    set hash $d(_hex_)
    catch {unset d o}
    return $hash
}


# CODEINFO BEGIN ######################################################
# info proc getUserName> Procedure to determine the name of the
#     current user

# arg> None
# return> User name

proc webSubmit::util::getUserName {} {
    # Create zero size temp file
    set tmpFile "/tmp/owner.[pid]"
    tclTouch $tmpFile

    # Retrieve owner of file
    set owner [file attributes $tmpFile -owner]

    # Delete file
    file delete -force $tmpFile

    return $owner
}


# CODEINFO BEGIN ######################################################
# info proc isUser> Procedure to determine if the current user matches
#     a specified user name

# arg user> User name to test
# return> 1 = True, 0 = False

proc webSubmit::util::isUser {user} {
    if {[getUserName] != "$user"} {
	return 0
    } else {
	return 1
    }
}


# CODEINFO BEGIN ######################################################
# info proc tclDate> procedure to return the current time and date
# arg> None
# return> Time and Date

proc webSubmit::util::tclDate {} {
    return [clock format [clock seconds]]
}


# CODEINFO BEGIN ######################################################
# info proc tclCat> Procedure to simulate UNIX cat for a file
# arg> fileName
# return> File contents

proc webSubmit::util::tclCat {fileName} {
    if {[file isfile $fileName] && [file readable $fileName]} {
	set fd [open $fileName r]
	set fileData [read $fd]
	close $fd
    } else {
	return -code error "File \"$fileName\" does not exist or is unreadable"
    }
    return $fileData
}

